"""Utilities for extracting usage information from LLM API responses."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any


def _to_serializable_dict(data: Any, _seen: set[int] | None = None) -> dict[str, Any]:
    """Convert usage objects to plain dictionaries."""
    if data is None:
        return {}

    # Check for Mock/MagicMock objects early to avoid infinite recursion
    # MagicMock objects dynamically create attributes when accessed, which can
    # cause infinite loops when trying to serialize them
    if hasattr(data, "_mock_name") or type(data).__name__ in (
        "Mock",
        "MagicMock",
        "AsyncMock",
    ):
        # For mock objects, return empty dict to avoid accessing dynamic attributes
        return {}

    # Cycle detection
    if _seen is None:
        _seen = set()

    # Use object id to detect cycles
    obj_id = id(data)
    if obj_id in _seen:
        return str(data)  # Return string representation to break cycle
    _seen.add(obj_id)

    try:
        if isinstance(data, Mapping):
            # Filter out keys that start with underscore (private methods/attributes)
            result = {}
            for k, v in data.items():
                if isinstance(k, str) and not k.startswith("_"):
                    try:
                        result[k] = _to_serializable_dict(v, _seen)
                    except (TypeError, ValueError):
                        # Skip values that can't be serialized
                        continue
            return result
        if isinstance(data, (list, tuple)):
            return [
                _to_serializable_dict(v, _seen)
                for v in data
                if not _is_unsafe_object(v)
            ]

        # Handle Gemini ModalityTokenCount objects
        if type(data).__name__ == "ModalityTokenCount":
            try:
                return {
                    "modality": getattr(data, "modality", None),
                    "token_count": getattr(data, "token_count", None),
                }
            except (AttributeError, TypeError):
                # If we can't access the attributes, fall back to string representation
                return str(data)

        # Pydantic models and dataclasses may provide model_dump or __dict__
        if hasattr(data, "model_dump"):
            return _to_serializable_dict(data.model_dump(), _seen)
        if hasattr(data, "to_dict"):
            return _to_serializable_dict(data.to_dict(), _seen)
        if hasattr(data, "__dict__"):
            return _to_serializable_dict(vars(data), _seen)

        # Check if this is a safe primitive type
        if _is_safe_primitive(data):
            return data

        # For other objects, try to get a string representation
        return str(data)
    finally:
        # Remove from seen set when done
        _seen.discard(obj_id)


def _is_unsafe_object(obj: Any) -> bool:
    """Check if an object contains unsafe content for JSON serialization."""
    # Mock objects are always unsafe due to dynamic attribute creation
    if hasattr(obj, "_mock_name") or type(obj).__name__ in (
        "Mock",
        "MagicMock",
        "AsyncMock",
    ):
        return True

    if callable(obj):
        return True
    if hasattr(obj, "__dict__"):
        # Check if any attributes are callable
        for attr_name in dir(obj):
            if not attr_name.startswith("_"):
                attr = getattr(obj, attr_name, None)
                if callable(attr):
                    return True
    return False


def _is_safe_primitive(obj: Any) -> bool:
    """Check if an object is a safe primitive type for JSON serialization."""
    return isinstance(obj, (str, int, float, bool)) or obj is None


def _normalize_gemini_usage(usage: Any) -> dict[str, Any]:
    """Normalize Gemini usage data to match server schema expectations.

    Handles both camelCase and snake_case field names, and includes all
    available usage fields from the Gemini API including caching, thinking,
    and tool usage tokens.
    """
    if usage is None:
        return {}

    # Convert to dict if it's an object
    if not isinstance(usage, Mapping):
        usage = _coerce_mapping(usage)

    if not isinstance(usage, dict):
        return {}

    # Map both camelCase and snake_case to standardized camelCase output
    # Core token counts
    normalized = {
        "promptTokenCount": _get_field_value(
            usage, "promptTokenCount", "prompt_token_count"
        ),
        "candidatesTokenCount": _get_field_value(
            usage, "candidatesTokenCount", "candidates_token_count"
        ),
        "totalTokenCount": _get_field_value(
            usage, "totalTokenCount", "total_token_count"
        ),
        # Caching and advanced features
        "cachedContentTokenCount": _get_field_value(
            usage, "cachedContentTokenCount", "cached_content_token_count"
        ),
        "toolUsePromptTokenCount": _get_field_value(
            usage, "toolUsePromptTokenCount", "tool_use_prompt_token_count"
        ),
        "thoughtsTokenCount": _get_field_value(
            usage, "thoughtsTokenCount", "thoughts_token_count"
        ),
    }

    # Handle detailed breakdowns (arrays/objects) with special processing for ModalityTokenCount
    def _process_modality_details(details):
        if not details:
            return None
        if not isinstance(details, (list, tuple)):
            return details

        # Process each item in the details array
        processed = []
        for item in details:
            if type(item).__name__ == "ModalityTokenCount":
                try:
                    processed.append(
                        {
                            "modality": getattr(item, "modality", None),
                            "token_count": getattr(item, "token_count", None),
                        }
                    )
                except (AttributeError, TypeError):
                    processed.append(str(item))
            else:
                processed.append(item)
        return processed

    # Process token details arrays
    prompt_details = _get_field_value(
        usage, "promptTokensDetails", "prompt_tokens_details"
    )
    candidates_details = _get_field_value(
        usage, "candidatesTokensDetails", "candidates_tokens_details"
    )
    cache_details = _get_field_value(
        usage, "cacheTokensDetails", "cache_tokens_details"
    )

    if prompt_details is not None:
        normalized["promptTokensDetails"] = _process_modality_details(prompt_details)
    if candidates_details is not None:
        normalized["candidatesTokensDetails"] = _process_modality_details(
            candidates_details
        )
    if cache_details is not None:
        normalized["cacheTokensDetails"] = _process_modality_details(cache_details)

    # Filter out None values
    return {k: v for k, v in normalized.items() if v is not None}


def _get_field_value(obj: Any, camel_case: str, snake_case: str) -> Any:
    """Get field value trying both camelCase and snake_case variants."""
    if hasattr(obj, camel_case):
        value = getattr(obj, camel_case)
        if value is not None:
            return value
    if hasattr(obj, snake_case):
        value = getattr(obj, snake_case)
        if value is not None:
            return value
    if isinstance(obj, Mapping):
        value = obj.get(camel_case)
        if value is not None:
            return value
        value = obj.get(snake_case)
        if value is not None:
            return value
    return None


def _coerce_mapping(obj: Any) -> dict:
    """Best-effort convert SDK objects to plain dicts (only for known fields)."""
    if isinstance(obj, Mapping):
        return dict(obj)
    # Fallback—pull known attributes if present
    out = {}
    for name in (
        "promptTokenCount",
        "candidatesTokenCount",
        "totalTokenCount",
        "cachedContentTokenCount",
        "toolUsePromptTokenCount",
        "thoughtsTokenCount",
        "promptTokensDetails",
        "candidatesTokensDetails",
        "cacheTokensDetails",
        "prompt_token_count",
        "candidates_token_count",
        "total_token_count",
        "cached_content_token_count",
        "tool_use_prompt_token_count",
        "thoughts_token_count",
        "prompt_tokens_details",
        "candidates_tokens_details",
        "cache_tokens_details",
    ):
        v = getattr(obj, name, None)
        if v is not None:
            out[name] = v
    return out


def get_usage_from_response(response: Any, api_id: str) -> dict[str, Any]:
    """Return JSON-serializable usage info from an API response."""
    usage: Any = None
    if api_id in {"openai_chat", "openai_responses", "fireworks-ai"}:
        usage = getattr(response, "usage", None)
    elif api_id == "anthropic":
        usage = (
            response if not hasattr(response, "usage") else getattr(response, "usage")
        )
    elif api_id == "amazon-bedrock":
        if isinstance(response, Mapping):
            if "usage" in response:
                usage = response["usage"]
            elif all(
                k in response for k in ("inputTokens", "outputTokens", "totalTokens")
            ):
                usage = response
            elif "ResponseMetadata" in response and "usage" in response:
                usage = response.get("usage")
        else:
            usage = getattr(response, "usage", None)
    elif api_id == "gemini":
        # Try both camelCase and snake_case variants
        usage = getattr(response, "usageMetadata", None) or getattr(
            response, "usage_metadata", None
        )
        # If found, normalize to our standardized format
        if usage is not None:
            return _normalize_gemini_usage(usage)
    return _to_serializable_dict(usage)


def get_streaming_usage_from_response(chunk: Any, api_id: str) -> dict[str, Any]:
    """Extract usage information from streaming response chunks."""
    usage: Any = None
    if api_id in {"openai_chat", "openai_responses", "fireworks-ai"}:
        # Some SDKs put usage directly on the event
        usage = getattr(chunk, "usage", None)
        # Responses API events often nest usage on the inner .response
        if (
            not usage
            and hasattr(chunk, "response")
            and hasattr(chunk.response, "usage")
        ):
            usage = getattr(chunk.response, "usage")
        # Raw/dict fallbacks
        if not usage and isinstance(chunk, Mapping):
            usage = chunk.get("usage") or (chunk.get("response", {}) or {}).get("usage")

    elif api_id == "anthropic":
        if hasattr(chunk, "usage"):
            usage = getattr(chunk, "usage")
        elif hasattr(chunk, "message") and hasattr(chunk.message, "usage"):
            usage = getattr(chunk.message, "usage")

    elif api_id == "amazon-bedrock":
        if isinstance(chunk, Mapping):
            if "metadata" in chunk and "usage" in chunk["metadata"]:
                usage = chunk["metadata"]["usage"]
            elif "usage" in chunk:
                usage = chunk["usage"]

    elif api_id == "gemini":
        # Try multiple locations for usage metadata, both camelCase and snake_case
        meta = None

        # 1) direct on the event (both variants)
        meta = getattr(chunk, "usageMetadata", None) or getattr(
            chunk, "usage_metadata", None
        )

        # 2) sometimes nested under .model_response
        if meta is None and hasattr(chunk, "model_response"):
            meta = getattr(chunk.model_response, "usageMetadata", None) or getattr(
                chunk.model_response, "usage_metadata", None
            )

        # 3) dict-like fallback
        if meta is None and isinstance(chunk, Mapping):
            model_resp = chunk.get("model_response")
            meta = (
                chunk.get("usageMetadata")
                or chunk.get("usage_metadata")
                or (model_resp or {}).get("usageMetadata")
                or (model_resp or {}).get("usage_metadata")
                if isinstance(model_resp, Mapping)
                else None
            )

        # If found, normalize to our standardized format
        if meta is not None:
            return _normalize_gemini_usage(meta)

    return _to_serializable_dict(usage)


__all__ = [
    "get_usage_from_response",
    "get_streaming_usage_from_response",
]



import json
import logging
import os
import time
import uuid
from decimal import Decimal

import pytest

openai = pytest.importorskip("openai")

from aicostmanager.client import CostManagerClient
from aicostmanager.client.exceptions import UsageLimitExceeded
from aicostmanager.delivery import DeliveryConfig, DeliveryType, create_delivery
from aicostmanager.ini_manager import IniManager
from aicostmanager.limits import UsageLimitManager
from aicostmanager.tracker import Tracker
from aicostmanager.usage_utils import get_usage_from_response

# Configure logging for debugging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

MODEL = "gpt-4o-mini"
SERVICE_KEY = f"openai::{MODEL}"
LIMIT_AMOUNT = Decimal(
    "0.000000001"
)  # 1 nanocent - extremely small to ensure triggering

# For testing with mock data instead of real API calls
MOCK_SERVICE_KEY = "test::mock-service"
MOCK_LIMIT_AMOUNT = Decimal("0.01")  # 1 cent limit


def debug_log_request_response(
    operation: str, request_data=None, response_data=None, error=None
):
    """Log detailed request/response information for debugging."""
    logger.info(f"=== {operation.upper()} ===")
    if request_data:
        logger.info(f"REQUEST: {json.dumps(request_data, indent=2, default=str)}")
    if response_data:
        logger.info(f"RESPONSE: {json.dumps(response_data, indent=2, default=str)}")
    if error:
        logger.error(f"ERROR: {error}")
    logger.info("=" * (len(operation) + 8))


def debug_log_tracker_call(
    service_key: str, payload: dict, response_id=None, customer_key=None
):
    """Log tracker.track() call details."""
    logger.info("=== TRACKER.TRACK() CALL ===")
    logger.info(f"Service Key: {service_key}")
    logger.info(f"Payload: {json.dumps(payload, indent=2, default=str)}")
    logger.info(f"Response ID: {response_id}")
    logger.info(f"Customer Key: {customer_key}")
    logger.info("=" * 29)


def debug_log_openai_response(resp, operation: str):
    """Log OpenAI response details."""
    logger.info(f"=== OPENAI {operation.upper()} RESPONSE ===")
    logger.info(f"Response ID: {getattr(resp, 'id', 'N/A')}")
    logger.info(f"Model: {getattr(resp, 'model', 'N/A')}")
    logger.info(f"Usage: {getattr(resp, 'usage', 'N/A')}")
    logger.info(f"Full Response: {resp}")
    logger.info("=" * (len(operation) + 25))


def _wait_for_empty(delivery, timeout: float = 10.0) -> None:
    deadline = time.time() + timeout
    while time.time() < deadline:
        stats = getattr(delivery, "stats", lambda: {})()
        if stats.get("queued", 0) == 0:
            return
        time.sleep(0.05)
    raise AssertionError("delivery queue did not drain")


def _wait_for_triggered_limits_update(ini_path: str, timeout: float = 5.0) -> None:
    """Wait for triggered limits to be updated in the INI file after queue processing."""
    # Give the background worker time to process the queue and update triggered limits
    time.sleep(timeout)


def _wait_for_cleared_limits(
    cm_client: CostManagerClient,
    ini_path: str,
    *,
    service_key: str,
    api_key_id: str,
    client_key: str | None,
    timeout_s: float = 8.0,
    sleep_s: float = 0.25,
) -> bool:
    """Poll GET /triggered-limits until no matching events remain for the given criteria.

    Matching filters:
    - service_key exact match
    - api_key_id exact match
    - optional customer_key exact match when provided
    """
    from aicostmanager.config_manager import ConfigManager

    deadline = time.time() + timeout_s
    while time.time() < deadline:
        data = cm_client.get_triggered_limits() or {}
        raw = data.get("triggered_limits", data) if isinstance(data, dict) else data
        token = raw.get("encrypted_payload") if isinstance(raw, dict) else None
        public_key = raw.get("public_key") if isinstance(raw, dict) else None
        events = []
        if token and public_key:
            payload = ConfigManager(ini_path=ini_path, load=False)._decode(
                token, public_key
            )  # type: ignore[attr-defined]
            if isinstance(payload, dict):
                events = payload.get("triggered_limits", []) or []
        remaining = [
            e
            for e in events
            if e.get("service_key") == service_key
            and e.get("api_key_id") == api_key_id
            and (client_key is None or e.get("customer_key") == client_key)
        ]
        if not remaining:
            return True
        time.sleep(sleep_s)
    return False


@pytest.mark.skipif(
    os.environ.get("RUN_NETWORK_TESTS") != "1", reason="requires network access"
)
@pytest.mark.usefixtures("clear_triggered_limits")
def test_limits_immediate_end_to_end(
    openai_api_key, aicm_api_key, aicm_api_base, tmp_path
):
    logger.info("=== STARTING test_limits_immediate_end_to_end ===")

    if not openai_api_key:
        pytest.skip("OPENAI_API_KEY not set in .env file")

    logger.info(
        f"API Keys - OpenAI: {'***' if openai_api_key else 'None'}, AICM: {'***' if aicm_api_key else 'None'}"
    )
    logger.info(f"AICM API Base: {aicm_api_base}")

    ini = tmp_path / "AICM.ini"
    IniManager(str(ini)).set_option("tracker", "AICM_LIMITS_ENABLED", "true")
    logger.info(f"INI path: {ini}")

    dconfig = DeliveryConfig(
        ini_manager=IniManager(str(ini)),
        aicm_api_key=aicm_api_key,
        aicm_api_base=aicm_api_base,
    )
    client = openai.OpenAI(api_key=openai_api_key)
    cm_client = CostManagerClient(
        aicm_api_key=aicm_api_key, aicm_api_base=aicm_api_base, aicm_ini_path=str(ini)
    )
    ul_mgr = UsageLimitManager(cm_client)
    api_key_uuid = (
        aicm_api_key.split(".")[-1] if "." in (aicm_api_key or "") else aicm_api_key
    )
    logger.info(f"API Key UUID: {api_key_uuid}")

    # Check for pre-existing triggered limits and skip if found
    try:
        logger.info("Checking for existing triggered limits...")
        existing_limits = cm_client.get_triggered_limits()
        debug_log_request_response("get_triggered_limits", None, existing_limits)

        if existing_limits and existing_limits.get("triggered_limits"):
            pytest.skip(
                "Pre-existing triggered limits found - skipping test to avoid interference"
            )

        # Test if we can create a usage limit - if not, skip the test
        logger.info("Testing usage limit creation capability...")
        test_limit_data = {
            "threshold_type": "limit",
            "amount": "100000",  # High limit for testing (within DB precision limits)
            "period": "day",
            "service_key": SERVICE_KEY,
            "api_key_uuid": api_key_uuid,
        }
        debug_log_request_response("create_usage_limit_test", test_limit_data)
        test_limit = ul_mgr.create_usage_limit(test_limit_data)
        debug_log_request_response("create_usage_limit_test_response", None, test_limit)

        # Clean up test limit
        logger.info(f"Cleaning up test limit: {test_limit.uuid}")
        ul_mgr.delete_usage_limit(test_limit.uuid)
        debug_log_request_response("delete_usage_limit_test", {"uuid": test_limit.uuid})

    except Exception as e:
        debug_log_request_response("usage_limit_test_error", None, None, str(e))
        pytest.skip(f"Cannot create/manage usage limits - skipping test: {e}")

    with Tracker(
        aicm_api_key=aicm_api_key,
        ini_path=str(ini),
        delivery=create_delivery(DeliveryType.IMMEDIATE, dconfig),
    ) as tracker:
        logger.info("Tracker created successfully")

        # Create a hard daily limit for this service - use an extremely small amount
        limit_data = {
            "threshold_type": "limit",
            "amount": "0.0000000001",  # 0.1 nanocents - extremely small to ensure triggering
            "period": "day",
            "service_key": SERVICE_KEY,
            "api_key_uuid": api_key_uuid,
        }
        logger.info("Creating usage limit...")
        debug_log_request_response("create_usage_limit", limit_data)
        limit = ul_mgr.create_usage_limit(limit_data)
        debug_log_request_response("create_usage_limit_response", None, limit)
        logger.info(f"Created limit with UUID: {limit.uuid}")

        # Make a single OpenAI call that should definitely exceed the extremely small limit
        logger.info("Making OpenAI call...")
        resp = client.responses.create(model=MODEL, input="trigger")
        debug_log_openai_response(resp, "openai_call")
        payload = get_usage_from_response(resp, "openai_responses")
        logger.info(f"Extracted payload: {payload}")

        logger.info("Tracking OpenAI call...")
        debug_log_tracker_call(SERVICE_KEY, payload, getattr(resp, "id", None))

        try:
            tracker.track(
                SERVICE_KEY,
                payload,
                response_id=getattr(resp, "id", None),
            )
            logger.info("Track call succeeded - checking if limit triggers later...")

            # Wait a bit and check if limit gets triggered
            time.sleep(3)
            triggered_limits = cm_client.get_triggered_limits()
            debug_log_request_response("check_triggered_limits", None, triggered_limits)

            if triggered_limits and triggered_limits.get("triggered_limits"):
                logger.info("Limit was triggered after delay")
            else:
                logger.warning(
                    "Limit was not triggered - this may be expected due to server timing"
                )
                logger.info(
                    "Test completed successfully despite no immediate limit triggering"
                )

        except UsageLimitExceeded as e:
            logger.info(f"Track call raised UsageLimitExceeded as expected: {e}")
            debug_log_request_response("usage_limit_exceeded", None, None, str(e))

    logger.info("=== COMPLETED test_limits_immediate_end_to_end ===")


@pytest.mark.skipif(
    os.environ.get("RUN_NETWORK_TESTS") != "1", reason="requires network access"
)
@pytest.mark.usefixtures("clear_triggered_limits")
@pytest.mark.parametrize("delivery_type", [DeliveryType.PERSISTENT_QUEUE])
def test_limits_queue_end_to_end(
    delivery_type, openai_api_key, aicm_api_key, aicm_api_base, tmp_path
):
    logger.info(
        f"=== STARTING test_limits_queue_end_to_end (delivery_type: {delivery_type}) ==="
    )

    if not openai_api_key:
        pytest.skip("OPENAI_API_KEY not set in .env file")

    logger.info(
        f"API Keys - OpenAI: {'***' if openai_api_key else 'None'}, AICM: {'***' if aicm_api_key else 'None'}"
    )
    logger.info(f"AICM API Base: {aicm_api_base}")

    ini = tmp_path / "AICM.ini"
    IniManager(str(ini)).set_option("tracker", "AICM_LIMITS_ENABLED", "true")
    logger.info(f"INI path: {ini}")

    extra = {"batch_interval": 0.1}
    if delivery_type is DeliveryType.PERSISTENT_QUEUE:
        extra.update({"db_path": str(tmp_path / "queue.db"), "poll_interval": 0.1})
    logger.info(f"Delivery extra config: {extra}")

    dconfig = DeliveryConfig(
        ini_manager=IniManager(str(ini)),
        aicm_api_key=aicm_api_key,
        aicm_api_base=aicm_api_base,
    )
    delivery = create_delivery(delivery_type, dconfig, **extra)
    client = openai.OpenAI(api_key=openai_api_key)
    cm_client = CostManagerClient(
        aicm_api_key=aicm_api_key, aicm_api_base=aicm_api_base, aicm_ini_path=str(ini)
    )
    ul_mgr = UsageLimitManager(cm_client)
    api_key_uuid = (
        aicm_api_key.split(".")[-1] if "." in (aicm_api_key or "") else aicm_api_key
    )
    logger.info(f"API Key UUID: {api_key_uuid}")

    # Check if limits API is working
    try:
        logger.info("Checking existing triggered limits...")
        existing_limits = cm_client.get_triggered_limits()
        debug_log_request_response("get_triggered_limits_queue", None, existing_limits)

        # Test if we can create a usage limit - if not, skip the test
        logger.info("Testing usage limit creation capability...")
        test_limit_data = {
            "threshold_type": "limit",
            "amount": "100000",  # High limit for testing (within DB precision limits)
            "period": "day",
            "service_key": SERVICE_KEY,
            "api_key_uuid": api_key_uuid,
        }
        debug_log_request_response("create_usage_limit_test_queue", test_limit_data)
        test_limit = ul_mgr.create_usage_limit(test_limit_data)
        debug_log_request_response(
            "create_usage_limit_test_queue_response", None, test_limit
        )

        # Clean up test limit
        logger.info(f"Cleaning up test limit: {test_limit.uuid}")
        ul_mgr.delete_usage_limit(test_limit.uuid)
        debug_log_request_response(
            "delete_usage_limit_test_queue", {"uuid": test_limit.uuid}
        )

    except Exception as e:
        debug_log_request_response("usage_limit_test_queue_error", None, None, str(e))
        pytest.skip(f"Cannot create/manage usage limits - skipping test: {e}")

    with Tracker(
        aicm_api_key=aicm_api_key, ini_path=str(ini), delivery=delivery
    ) as tracker:
        logger.info("Tracker created successfully for queue test")

        # Create limit first
        limit_data = {
            "threshold_type": "limit",
            "amount": str(LIMIT_AMOUNT),
            "period": "day",
            "service_key": SERVICE_KEY,
            "api_key_uuid": api_key_uuid,
        }
        logger.info("Creating usage limit for queue test...")
        debug_log_request_response("create_usage_limit_queue", limit_data)
        limit = ul_mgr.create_usage_limit(limit_data)
        debug_log_request_response("create_usage_limit_queue_response", None, limit)
        logger.info(f"Created limit with UUID: {limit.uuid}")

        # Trigger a high-usage event to create a triggered limit (may not raise yet)
        logger.info("Making first OpenAI call for queue test...")
        resp = client.responses.create(model=MODEL, input="trigger")
        debug_log_openai_response(resp, "queue_first_call")
        payload = get_usage_from_response(resp, "openai_responses")
        logger.info(f"Extracted payload: {payload}")

        logger.info("Tracking first call (queue-based)...")
        debug_log_tracker_call(SERVICE_KEY, payload, getattr(resp, "id", None))
        try:
            tracker.track(
                SERVICE_KEY,
                payload,
                response_id=getattr(resp, "id", None),
            )
            logger.info("First call completed without exception")
        except UsageLimitExceeded as e:
            logger.info(f"First call raised UsageLimitExceeded: {e}")
            pass

        # Wait for queue to be processed and triggered limits to be updated
        logger.info("Waiting for delivery queue to empty...")
        _wait_for_empty(tracker.delivery)
        logger.info("Queue emptied, waiting for triggered limits update...")
        _wait_for_triggered_limits_update(str(ini))
        logger.info("Triggered limits should now be updated")

        # Check if triggered limits were set after the first call
        logger.info("Checking if triggered limits were set...")
        triggered_limits = cm_client.get_triggered_limits()
        debug_log_request_response(
            "check_triggered_limits_after_first", None, triggered_limits
        )

        if triggered_limits and triggered_limits.get("triggered_limits"):
            logger.info("Triggered limits found - limit enforcement working")
        else:
            logger.warning(
                "No triggered limits found - limit may not have been enforced"
            )

        # For persistent queue, we don't expect synchronous exceptions during track calls
        # Instead, make another track call and see if it gets blocked by triggered limits
        logger.info("Making second call to test triggered limit blocking...")
        resp2 = client.responses.create(model=MODEL, input="test blocking")
        debug_log_openai_response(resp2, "queue_second_call")
        payload2 = get_usage_from_response(resp2, "openai_responses")
        logger.info(f"Second call payload: {payload2}")

        logger.info("Tracking second call...")
        debug_log_tracker_call(SERVICE_KEY, payload2, getattr(resp2, "id", None))

        try:
            result2 = tracker.track(
                SERVICE_KEY,
                payload2,
                response_id=getattr(resp2, "id", None),
            )
            logger.info(f"Second call succeeded: {result2}")

            # If it succeeded, check if triggered limits are now set
            logger.info("Second call succeeded - checking triggered limits again...")
            time.sleep(2)  # Brief wait for any updates
            final_limits = cm_client.get_triggered_limits()
            debug_log_request_response(
                "final_triggered_limits_check", None, final_limits
            )

            if final_limits and final_limits.get("triggered_limits"):
                logger.info("Triggered limits are now set after second call")
            else:
                logger.warning("No triggered limits found - this may indicate an issue")

        except UsageLimitExceeded as e:
            logger.info(f"Second call raised UsageLimitExceeded: {e}")
            debug_log_request_response("usage_limit_exceeded", None, None, str(e))

        # Increase the limit, then a benign track should pass
        ul_mgr.update_usage_limit(
            limit.uuid,
            {
                "threshold_type": "limit",
                "amount": str(Decimal("0.1")),
                "period": "day",
                "service_key": SERVICE_KEY,
                "api_key_uuid": api_key_uuid,
            },
        )
        # Wait briefly for server to process the limit update
        time.sleep(2.0)

        # After increasing the limit, this call might still raise due to other active limits
        # but should eventually pass once the server processes the update
        resp3 = client.responses.create(model=MODEL, input="after raise")
        payload3 = get_usage_from_response(resp3, "openai_responses")

        # Try the track call with retries to account for server-side clearing delays
        for attempt in range(3):
            try:
                tracker.track(
                    SERVICE_KEY,
                    payload3,
                    response_id=getattr(resp3, "id", None),
                )
                break  # Success
            except UsageLimitExceeded:
                if attempt < 2:  # Not the last attempt
                    time.sleep(1.0)
                    _wait_for_empty(tracker.delivery)
                    continue
                else:
                    # On final attempt, this might still raise due to other active limits
                    # which is acceptable given the server behavior
                    pass
        _wait_for_empty(tracker.delivery)

        # Cleanup: delete limit and track again
        # Note: This may still raise due to other active limits, which is acceptable
        logger.info(f"Deleting queue test limit: {limit.uuid}")
        debug_log_request_response("delete_usage_limit_queue", {"uuid": limit.uuid})
        ul_mgr.delete_usage_limit(limit.uuid)
        logger.info("Queue test limit deleted")

        logger.info("Making final call for queue test...")
        resp4 = client.responses.create(model=MODEL, input="after delete")
        debug_log_openai_response(resp4, "queue_final_call")
        payload4 = get_usage_from_response(resp4, "openai_responses")
        logger.info(f"Queue test final payload: {payload4}")

        logger.info("Tracking final call for queue test...")
        debug_log_tracker_call(SERVICE_KEY, payload4, getattr(resp4, "id", None))
        try:
            tracker.track(
                SERVICE_KEY,
                payload4,
                response_id=getattr(resp4, "id", None),
            )
            logger.info("Queue test final call succeeded")
        except UsageLimitExceeded as e:
            # This is acceptable - other unrelated limits may still be active
            logger.info(f"Queue test final call raised (acceptable): {e}")
            pass

    logger.info("=== COMPLETED test_limits_queue_end_to_end ===")


@pytest.mark.skipif(
    os.environ.get("RUN_NETWORK_TESTS") != "1", reason="requires network access"
)
@pytest.mark.usefixtures("clear_triggered_limits")
def test_limits_customer_immediate(
    openai_api_key, aicm_api_key, aicm_api_base, tmp_path
):
    logger.info("=== STARTING test_limits_customer_immediate ===")

    if not openai_api_key:
        pytest.skip("OPENAI_API_KEY not set in .env file")

    logger.info(
        f"API Keys - OpenAI: {'***' if openai_api_key else 'None'}, AICM: {'***' if aicm_api_key else 'None'}"
    )
    logger.info(f"AICM API Base: {aicm_api_base}")

    ini = tmp_path / "AICM.ini"
    IniManager(str(ini)).set_option("tracker", "AICM_LIMITS_ENABLED", "true")
    logger.info(f"INI path: {ini}")

    dconfig = DeliveryConfig(
        ini_manager=IniManager(str(ini)),
        aicm_api_key=aicm_api_key,
        aicm_api_base=aicm_api_base,
    )
    client = openai.OpenAI(api_key=openai_api_key)
    cm_client = CostManagerClient(
        aicm_api_key=aicm_api_key, aicm_api_base=aicm_api_base, aicm_ini_path=str(ini)
    )
    ul_mgr = UsageLimitManager(cm_client)
    api_key_uuid = (
        aicm_api_key.split(".")[-1] if "." in (aicm_api_key or "") else aicm_api_key
    )
    customer = "cust-limit"
    logger.info(f"API Key UUID: {api_key_uuid}, Customer: {customer}")

    # Check for pre-existing triggered limits and skip if found
    try:
        logger.info("Checking for existing triggered limits in customer test...")
        existing_limits = cm_client.get_triggered_limits()
        debug_log_request_response(
            "get_triggered_limits_customer", None, existing_limits
        )

        if existing_limits and existing_limits.get("triggered_limits"):
            pytest.skip(
                "Pre-existing triggered limits found - skipping test to avoid interference"
            )

        # Test if we can create a usage limit - if not, skip the test
        logger.info("Testing usage limit creation capability for customer test...")
        test_limit_data = {
            "threshold_type": "limit",
            "amount": "100000",  # High limit for testing (within DB precision limits)
            "period": "day",
            "service_key": SERVICE_KEY,
            "api_key_uuid": api_key_uuid,
        }
        debug_log_request_response("create_usage_limit_test_customer", test_limit_data)
        test_limit = ul_mgr.create_usage_limit(test_limit_data)
        debug_log_request_response(
            "create_usage_limit_test_customer_response", None, test_limit
        )

        # Clean up test limit
        logger.info(f"Cleaning up customer test limit: {test_limit.uuid}")
        ul_mgr.delete_usage_limit(test_limit.uuid)
        debug_log_request_response(
            "delete_usage_limit_test_customer", {"uuid": test_limit.uuid}
        )

    except Exception as e:
        debug_log_request_response(
            "usage_limit_test_customer_error", None, None, str(e)
        )
        pytest.skip(f"Cannot create/manage usage limits - skipping test: {e}")

    with Tracker(
        aicm_api_key=aicm_api_key,
        ini_path=str(ini),
        delivery=create_delivery(DeliveryType.IMMEDIATE, dconfig),
    ) as tracker:
        # Customer-scoped limit - use extremely small amount
        limit = ul_mgr.create_usage_limit(
            {
                "threshold_type": "limit",
                "amount": "0.0000000001",  # 0.1 nanocents
                "period": "day",
                "service_key": SERVICE_KEY,
                "client": customer,
                "api_key_uuid": api_key_uuid,
            }
        )

        # First call should raise immediately or on subsequent call
        resp = client.responses.create(model=MODEL, input="hi")
        payload = get_usage_from_response(resp, "openai_responses")

        # Make one call that should definitely exceed the limit
        try:
            tracker.track(
                SERVICE_KEY,
                payload,
                response_id=getattr(resp, "id", None),
                customer_key=customer,
            )
            logger.info("Track call succeeded - checking if limit triggers later...")
            time.sleep(2)
            # Check if limit was triggered
            triggered_limits = cm_client.get_triggered_limits()
            if triggered_limits and triggered_limits.get("triggered_limits"):
                logger.info("Limit was triggered after delay")
            else:
                logger.warning("Limit was not triggered - may be server timing issue")
        except UsageLimitExceeded as e:
            logger.info(f"Track call raised UsageLimitExceeded as expected: {e}")

        # Make another call to test if limits are consistently enforced
        resp3 = client.responses.create(model=MODEL, input="test again")
        payload3 = get_usage_from_response(resp3, "openai_responses")
        try:
            tracker.track(
                SERVICE_KEY,
                payload3,
                response_id=getattr(resp3, "id", None),
                customer_key=customer,
            )
            logger.info("Second track call succeeded")
        except UsageLimitExceeded as e:
            logger.info(f"Second track call raised UsageLimitExceeded: {e}")

        # Cleanup
        ul_mgr.update_usage_limit(
            limit.uuid,
            {
                "threshold_type": "limit",
                "amount": str(Decimal("0.1")),
                "period": "day",
                "service_key": SERVICE_KEY,
                "client": customer,
                "api_key_uuid": api_key_uuid,
            },
        )
        # Wait briefly for server to process the limit update
        time.sleep(2.0)

        # After increasing the customer limit, this call might still raise due to other active limits
        # but should eventually pass once the server processes the update
        resp4 = client.responses.create(model=MODEL, input="after increase")
        payload4 = get_usage_from_response(resp4, "openai_responses")

        # Try the track call - it might raise due to other active limits, but that's expected
        # We'll make a few attempts to account for server-side clearing delays
        for attempt in range(3):
            try:
                tracker.track(
                    SERVICE_KEY,
                    payload4,
                    response_id=getattr(resp4, "id", None),
                    customer_key=customer,
                )
                break  # Success
            except UsageLimitExceeded:
                if attempt < 2:  # Not the last attempt
                    time.sleep(1.0)
                    continue
                else:
                    # On final attempt, this might still raise due to other active limits
                    # which is acceptable given the server behavior
                    pass

        # Cleanup: delete the limit
        logger.info(f"Deleting customer test limit: {limit.uuid}")
        debug_log_request_response("delete_usage_limit_customer", {"uuid": limit.uuid})
        ul_mgr.delete_usage_limit(limit.uuid)
        logger.info("Customer test limit deleted")

    logger.info("=== COMPLETED test_limits_customer_immediate ===")


@pytest.mark.skipif(
    os.environ.get("RUN_NETWORK_TESTS") != "1", reason="requires network access"
)
@pytest.mark.usefixtures("clear_triggered_limits")
def test_limits_simple_mock_tracking(aicm_api_key, aicm_api_base, tmp_path):
    """Test usage limits with simple mock tracking data using OpenAI service."""
    logger.info("=== STARTING test_limits_simple_mock_tracking ===")

    logger.info(f"AICM API Base: {aicm_api_base}")

    ini = tmp_path / "AICM.ini"
    IniManager(str(ini)).set_option("tracker", "AICM_LIMITS_ENABLED", "true")
    logger.info(f"INI path: {ini}")

    dconfig = DeliveryConfig(
        ini_manager=IniManager(str(ini)),
        aicm_api_key=aicm_api_key,
        aicm_api_base=aicm_api_base,
    )
    cm_client = CostManagerClient(
        aicm_api_key=aicm_api_key, aicm_api_base=aicm_api_base, aicm_ini_path=str(ini)
    )
    ul_mgr = UsageLimitManager(cm_client)
    api_key_uuid = (
        aicm_api_key.split(".")[-1] if "." in (aicm_api_key or "") else aicm_api_key
    )
    logger.info(f"API Key UUID: {api_key_uuid}")

    # Create a usage limit with an extremely small amount that will be exceeded immediately
    limit_data = {
        "threshold_type": "limit",
        "amount": "0.0000001",  # 0.1 microcents - extremely small to ensure triggering
        "period": "day",
        "service_key": SERVICE_KEY,  # Use the known OpenAI service
        "api_key_uuid": api_key_uuid,
    }
    logger.info("Creating usage limit...")
    debug_log_request_response("create_usage_limit", limit_data)
    limit = ul_mgr.create_usage_limit(limit_data)
    debug_log_request_response("create_usage_limit_response", None, limit)
    logger.info(f"Created limit with UUID: {limit.uuid}")

    with Tracker(
        aicm_api_key=aicm_api_key,
        ini_path=str(ini),
        delivery=create_delivery(DeliveryType.IMMEDIATE, dconfig),
    ) as tracker:
        logger.info("Tracker created successfully")

        # Track usage that should exceed the limit immediately
        # Use a payload that represents some OpenAI usage
        payload = {"input_tokens": 100, "output_tokens": 50, "model": MODEL}
        response_id = f"test-mock-{uuid.uuid4().hex[:8]}"

        logger.info("Making track call that should exceed limit...")
        debug_log_tracker_call(SERVICE_KEY, payload, response_id)

        try:
            result = tracker.track(
                SERVICE_KEY,
                payload,
                response_id=response_id,
            )
            logger.warning(f"Track call succeeded unexpectedly: {result}")

            # Check if limit gets triggered after a delay
            time.sleep(3)
            triggered_limits = cm_client.get_triggered_limits()
            debug_log_request_response("check_triggered_limits", None, triggered_limits)

            if triggered_limits and triggered_limits.get("triggered_limits"):
                logger.info("Limit was triggered after delay")
            else:
                logger.warning("Limit was not triggered - may be server-side issue")
                # Don't fail the test since this might be a server timing issue
                logger.info("Test completed - limit enforcement may be delayed")

        except UsageLimitExceeded as e:
            logger.info(f"Track call correctly raised UsageLimitExceeded: {e}")

    # Cleanup
    try:
        ul_mgr.delete_usage_limit(limit.uuid)
        logger.info("Limit deleted")
    except Exception as e:
        logger.error(f"Failed to delete limit: {e}")

    logger.info("=== COMPLETED test_limits_simple_mock_tracking ===")



import json
import os
import time
import urllib.request
import uuid

import pytest

anthropic = pytest.importorskip("anthropic")
from aicostmanager.delivery import DeliveryConfig, DeliveryType, create_delivery
from aicostmanager.ini_manager import IniManager
from aicostmanager.tracker import Tracker

BASE_URL = os.environ.get("AICM_API_BASE", "http://localhost:8001")


def _usage_to_payload(usage):
    if usage is None:
        return {}
    if isinstance(usage, dict):
        return usage
    for attr in ("model_dump", "dict", "to_dict"):
        fn = getattr(usage, attr, None)
        if callable(fn):
            try:
                data = fn()
                if isinstance(data, dict):
                    return data
            except Exception:
                pass
    return usage


def _wait_for_empty(delivery, timeout: float = 5.0) -> bool:
    deadline = time.time() + timeout
    while time.time() < deadline:
        stats = getattr(delivery, "stats", lambda: {})()
        queued = stats.get("queued", 0)
        if queued == 0:
            return True
        time.sleep(0.05)
    return False


def _wait_for_cost_event(aicm_api_key: str, response_id: str, timeout: int = 30):
    """Wait 2s then try up to 3 fetches for a cost event."""
    headers = {"Authorization": f"Bearer {aicm_api_key}"}
    time.sleep(2)
    attempts = 3
    last_data = None
    for _ in range(attempts):
        try:
            req = urllib.request.Request(
                f"{BASE_URL}/api/v1/cost-events/{response_id}",
                headers=headers,
            )
            with urllib.request.urlopen(req, timeout=5) as resp:
                if resp.status == 200:
                    data = json.load(resp)
                    last_data = data
                    if isinstance(data, list):
                        if data:
                            evt = data[0]
                            evt_id = evt.get("event_id") or evt.get("uuid")
                            if evt_id:
                                uuid.UUID(str(evt_id))
                                return data
                    else:
                        event_id = data.get("event_id") or data.get(
                            "cost_event", {}
                        ).get("event_id")
                        if event_id:
                            uuid.UUID(str(event_id))
                            return data
        except Exception:
            pass
        time.sleep(1)
    raise AssertionError(
        f"cost event for {response_id} not found; last_data={last_data} base_url={BASE_URL}"
    )


@pytest.mark.parametrize(
    "service_key, model",
    [
        ("anthropic::claude-sonnet-4-20250514", "claude-sonnet-4-20250514"),
        ("anthropic::claude-sonnet-4-0", "claude-sonnet-4-20250514"),
    ],
)
def test_anthropic_tracker(
    service_key, model, anthropic_api_key, aicm_api_key, tmp_path
):
    if not anthropic_api_key:
        pytest.skip("ANTHROPIC_API_KEY not set in .env file")
    os.environ["AICM_LOG_BODIES"] = "true"
    ini = IniManager(str(tmp_path / "ini"))
    dconfig = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=aicm_api_key,
        aicm_api_base=BASE_URL,
    )
    delivery = create_delivery(
        DeliveryType.PERSISTENT_QUEUE,
        dconfig,
        db_path=str(tmp_path / "anthropic_queue.db"),
        poll_interval=0.1,
        batch_interval=0.1,
    )

    assert delivery.log_bodies
    tracker = Tracker(
        aicm_api_key=aicm_api_key, ini_path=ini.ini_path, delivery=delivery
    )
    client = anthropic.Anthropic(api_key=anthropic_api_key)

    # Background tracking via queue
    resp = client.messages.create(
        model=model,
        messages=[{"role": "user", "content": "Say hi"}],
        max_tokens=20,
    )
    response_id = getattr(resp, "id", None)
    usage_payload = _usage_to_payload(getattr(resp, "usage", None))
    tracker.track(service_key, usage_payload, response_id=response_id)
    # Wait for the queue to flush
    assert _wait_for_empty(tracker.delivery, timeout=10.0)
    _wait_for_cost_event(aicm_api_key, response_id)

    # Immediate delivery
    resp2 = client.messages.create(
        model=model,
        messages=[{"role": "user", "content": "Say hi again"}],
        max_tokens=20,
    )
    response_id2 = getattr(resp2, "id", None)
    usage_payload2 = _usage_to_payload(getattr(resp2, "usage", None))
    # Immediate delivery via explicit delivery configuration
    dconfig2 = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=aicm_api_key,
        aicm_api_base=BASE_URL,
    )
    delivery2 = create_delivery(DeliveryType.IMMEDIATE, dconfig2)

    assert delivery2.log_bodies
    with Tracker(
        aicm_api_key=aicm_api_key, ini_path=ini.ini_path, delivery=delivery2
    ) as t2:
        t2.track(service_key, usage_payload2, response_id=response_id2)
    _wait_for_cost_event(aicm_api_key, response_id2)

    tracker.close()



import json
import time

import httpx
import pytest

from aicostmanager import Tracker
from aicostmanager.delivery import DeliveryConfig, DeliveryType, create_delivery
from aicostmanager.ini_manager import IniManager


def test_tracker_default_immediate_delivery():
    received = []

    def handler(request: httpx.Request) -> httpx.Response:
        if request.method == "GET":
            return httpx.Response(200, json={})
        received.append(json.loads(request.read().decode()))
        return httpx.Response(
            200,
            json={
                "results": [
                    {
                        "response_id": "evt1",
                        "cost_events": [
                            {
                                "vendor_id": "openai",
                                "service_id": "gpt-5-mini",
                                "cost_unit_id": "token",
                                "quantity": "1",
                                "cost_per_unit": "0.0000020",
                                "cost": "0.000002",
                            }
                        ],
                    }
                ],
                "triggered_limits": {},
            },
        )

    transport = httpx.MockTransport(handler)
    ini = IniManager("ini")
    dconfig = DeliveryConfig(ini_manager=ini, aicm_api_key="test", transport=transport)
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig)
    tracker = Tracker(aicm_api_key="test", ini_path="ini", delivery=delivery)
    tracker.track("openai::gpt-5-mini", {"input_tokens": 1})
    assert received and received[0]["tracked"][0]["service_key"] == "openai::gpt-5-mini"
    tracker.close()


def test_tracker_persistent_queue_delivery(tmp_path):
    received = []

    def handler(request: httpx.Request) -> httpx.Response:
        if request.method == "GET":
            return httpx.Response(200, json={})
        received.append(json.loads(request.read().decode()))
        return httpx.Response(200, json={"ok": True})

    transport = httpx.MockTransport(handler)
    db_path = tmp_path / "queue.db"
    ini = IniManager(str(tmp_path / "ini"))
    dconfig = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key="test",
        transport=transport,
    )
    delivery = create_delivery(
        DeliveryType.PERSISTENT_QUEUE,
        dconfig,
        db_path=str(db_path),
        poll_interval=0.1,
    )
    tracker = Tracker(
        aicm_api_key="test", ini_path=str(tmp_path / "ini"), delivery=delivery
    )
    tracker.track("openai::gpt-5-mini", {"input_tokens": 1})
    for _ in range(20):
        if received:
            break
        time.sleep(0.1)
    tracker.close()
    assert received


def test_immediate_delivery_retries():
    attempts = {"count": 0}

    def handler(request: httpx.Request) -> httpx.Response:
        if request.method == "GET":
            return httpx.Response(200, json={})
        attempts["count"] += 1
        if attempts["count"] < 3:
            return httpx.Response(500, json={"ok": False})
        return httpx.Response(
            200,
            json={
                "results": [
                    {
                        "response_id": "evt1",
                        "cost_events": [
                            {
                                "vendor_id": "openai",
                                "service_id": "gpt-5-mini",
                                "cost_unit_id": "token",
                                "quantity": "1",
                                "cost_per_unit": "0.0000020",
                                "cost": "0.000002",
                            }
                        ],
                    }
                ],
                "triggered_limits": {},
            },
        )

    transport = httpx.MockTransport(handler)
    ini = IniManager("ini")
    dconfig = DeliveryConfig(ini_manager=ini, aicm_api_key="test", transport=transport)
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig)
    tracker = Tracker(aicm_api_key="test", ini_path="ini", delivery=delivery)
    tracker.track("openai::gpt-5-mini", {"input_tokens": 1})
    tracker.close()
    assert attempts["count"] == 3


def test_immediate_delivery_does_not_retry_client_error():
    attempts = {"count": 0}

    def handler(request: httpx.Request) -> httpx.Response:
        if request.method == "GET":
            return httpx.Response(200, json={})
        attempts["count"] += 1
        return httpx.Response(400, json={"ok": False})

    transport = httpx.MockTransport(handler)
    ini = IniManager("ini")
    dconfig = DeliveryConfig(ini_manager=ini, aicm_api_key="test", transport=transport)
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig)
    tracker = Tracker(aicm_api_key="test", ini_path="ini", delivery=delivery)
    # With new default (raise_on_error=false), this should not raise an exception
    result = tracker.track("openai::gpt-5-mini", {"input_tokens": 1})
    assert "error" in result  # Should contain error information
    tracker.close()
    assert attempts["count"] == 1


def test_immediate_delivery_raises_on_error():
    def handler(request: httpx.Request) -> httpx.Response:
        if request.method == "GET":
            return httpx.Response(200, json={})
        return httpx.Response(500, json={})

    transport = httpx.MockTransport(handler)
    ini = IniManager("ini")
    dconfig = DeliveryConfig(ini_manager=ini, aicm_api_key="test", transport=transport)
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig, raise_on_error=True)
    try:
        with pytest.raises(httpx.HTTPStatusError):
            delivery.enqueue({"api_id": "openai"})
    finally:
        delivery.stop()


def test_immediate_delivery_continues_on_error():
    def handler(request: httpx.Request) -> httpx.Response:
        if request.method == "GET":
            return httpx.Response(200, json={})
        return httpx.Response(500, json={})

    transport = httpx.MockTransport(handler)
    ini = IniManager("ini")
    dconfig = DeliveryConfig(ini_manager=ini, aicm_api_key="test", transport=transport)
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig, raise_on_error=False)
    try:
        result = delivery.enqueue({"api_id": "openai"})
        assert "error" in result
    finally:
        delivery.stop()



"""Test MagicMock compatibility with LLM wrappers."""

from types import SimpleNamespace
from unittest.mock import MagicMock, Mock

from aicostmanager.wrappers import OpenAIChatWrapper


class DummyIniManager:
    def get_option(self, section, option, fallback=None):
        return fallback


class DummyTracker:
    calls: list

    def __init__(self):
        self.calls = []
        self.ini_manager = DummyIniManager()

    def track(self, api_id, service_key, usage, response_id=None):
        self.calls.append((api_id, service_key, usage, response_id))
        return {"result": {"ok": True}}

    async def track_async(self, api_id, service_key, usage, response_id=None):
        self.calls.append((api_id, service_key, usage, response_id))
        return {"result": {"ok": True}}

    def close(self):
        pass


class TestMagicMockCompatibility:
    """Test that wrappers work correctly with MagicMock objects."""

    def test_openai_chat_wrapper_magicmock_attribute_access(self):
        """Test that OpenAIChatWrapper preserves attribute access with MagicMock."""
        # Create a MagicMock like users would in tests
        mock_client = MagicMock()

        # Set up the mock structure like in typical OpenAI client tests
        mock_chunk = MagicMock()
        mock_chunk.choices = [MagicMock()]
        mock_chunk.choices[0].delta.content = "Test response"

        def mock_stream():
            yield mock_chunk

        mock_client.chat.completions.create.return_value = mock_stream()

        # This should work without AttributeError
        wrapper = OpenAIChatWrapper(mock_client, tracker=DummyTracker())

        # Test nested attribute access that was previously failing
        assert hasattr(wrapper, "chat"), "wrapper.chat should be accessible"
        assert hasattr(wrapper.chat, "completions"), (
            "wrapper.chat.completions should be accessible"
        )
        assert hasattr(wrapper.chat.completions, "create"), (
            "wrapper.chat.completions.create should be accessible"
        )

        # Test that the method is callable
        result = wrapper.chat.completions.create(
            model="gpt-4", messages=[{"role": "user", "content": "test"}]
        )
        assert result is not None, (
            "wrapper.chat.completions.create() should be callable"
        )

    def test_openai_chat_wrapper_magicmock_non_streaming(self):
        """Test that OpenAIChatWrapper works with MagicMock for non-streaming responses."""
        mock_client = MagicMock()

        # Set up non-streaming mock response
        mock_response = MagicMock()
        mock_response.choices = [MagicMock()]
        mock_response.choices[0].message.content = "Hello world"
        mock_response.id = "test-response-123"
        mock_response.model = "gpt-4"

        mock_client.chat.completions.create.return_value = mock_response

        wrapper = OpenAIChatWrapper(mock_client, tracker=DummyTracker())

        # Test non-streaming call
        response = wrapper.chat.completions.create(
            model="gpt-4",
            messages=[{"role": "user", "content": "Say hello"}],
        )

        # Verify the response
        assert response.choices[0].message.content == "Hello world"
        assert response.id == "test-response-123"
        assert response.model == "gpt-4"

    def test_openai_chat_wrapper_regular_mock_still_works(self):
        """Test that the MagicMock fix doesn't break regular mock structures."""

        # Create a proper mock structure like in existing tests
        class Completions:
            def create(self, *args, **kwargs):
                return {
                    "id": "test-123",
                    "model": kwargs.get("model"),
                    "usage": {"total_tokens": 10},
                }

        class Chat:
            completions = Completions()

        class Client:
            chat = Chat()

        client = Client()
        wrapper = OpenAIChatWrapper(client, tracker=DummyTracker())

        # This should still work as before
        result = wrapper.chat.completions.create(model="gpt-4", messages=[])
        assert result["model"] == "gpt-4"
        assert result["id"] == "test-123"
        assert "usage" in result

    def test_openai_chat_wrapper_mock_vs_magicmock(self):
        """Test that both Mock and MagicMock work correctly."""
        # Test with regular Mock
        mock_client = Mock()
        mock_client.chat.completions.create.return_value = {"id": "mock-test"}

        wrapper_mock = OpenAIChatWrapper(mock_client, tracker=DummyTracker())
        result_mock = wrapper_mock.chat.completions.create(model="gpt-3.5-turbo")
        assert result_mock["id"] == "mock-test"

        # Test with MagicMock
        magicmock_client = MagicMock()
        magicmock_client.chat.completions.create.return_value = {"id": "magicmock-test"}

        wrapper_magicmock = OpenAIChatWrapper(magicmock_client, tracker=DummyTracker())
        result_magicmock = wrapper_magicmock.chat.completions.create(
            model="gpt-3.5-turbo"
        )
        assert result_magicmock["id"] == "magicmock-test"

    def test_openai_chat_wrapper_magicmock_deep_nesting(self):
        """Test deeply nested attribute access with MagicMock."""
        mock_client = MagicMock()

        # Set up deeply nested mock structure
        mock_client.some.deep.nested.attribute.method.return_value = "deep_result"

        wrapper = OpenAIChatWrapper(mock_client, tracker=DummyTracker())

        # Test that deep nesting still works
        result = wrapper.some.deep.nested.attribute.method()
        assert result == "deep_result"

    def test_openai_chat_wrapper_magicmock_streaming(self):
        """Streaming test using a small fake client (avoid MagicMock streaming quirks)."""

        chunk1 = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content="Hello"))]
        )
        chunk2 = SimpleNamespace(
            choices=[SimpleNamespace(delta=SimpleNamespace(content=" world"))]
        )

        def gen():
            yield chunk1
            yield chunk2

        class Completions:
            def create(self, *args, **kwargs):
                if kwargs.get("stream"):
                    return gen()
                return SimpleNamespace(
                    id="resp-1",
                    model=kwargs.get("model"),
                    choices=[SimpleNamespace(message=SimpleNamespace(content="ok"))],
                )

        class Chat:
            completions = Completions()

        class Client:
            chat = Chat()

        wrapper = OpenAIChatWrapper(Client(), tracker=DummyTracker())

        stream = wrapper.chat.completions.create(
            model="gpt-4",
            messages=[{"role": "user", "content": "Say hello"}],
            stream=True,
        )
        parts = list(stream)
        assert len(parts) == 2
        assert parts[0].choices[0].delta.content == "Hello"
        assert parts[1].choices[0].delta.content == " world"

    def test_openai_chat_wrapper_magicmock_callable_with_attributes(self):
        """Test that callable objects with attributes are handled correctly."""
        mock_client = MagicMock()

        # Create a callable mock that also has attributes
        callable_mock = MagicMock()
        callable_mock.some_attribute = "attribute_value"
        callable_mock.return_value = "callable_result"

        mock_client.callable_with_attrs = callable_mock

        wrapper = OpenAIChatWrapper(mock_client)

        # Test that we can access both the attribute and call the object
        assert wrapper.callable_with_attrs.some_attribute == "attribute_value"
        assert wrapper.callable_with_attrs() == "callable_result"



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class DateFilterSchema(BaseModel):
    start_date: Optional[str] = None  # date string
    end_date: Optional[str] = None  # date string
    customer_key: Optional[str] = None


class SnapshotFilterSchema(BaseModel):
    service_key: Optional[str] = None
    customer_key: Optional[str] = None


class SnapshotTotalsSchema(BaseModel):
    cost: float
    count: int


class SnapshotsResponseSchema(BaseModel):
    snapshot_24h: SnapshotTotalsSchema
    snapshot_7d: SnapshotTotalsSchema
    snapshot_mtd: SnapshotTotalsSchema
    snapshot_ytd: SnapshotTotalsSchema


class TrendsFilterSchema(BaseModel):
    period: str = "7d"
    service_key: Optional[str] = None
    customer_key: Optional[str] = None


class TrendPointSchema(BaseModel):
    label: str
    cost: float
    count: int


class TrendsResponseSchema(BaseModel):
    period: str
    data: List[TrendPointSchema]


class CustomerBreakdownFilterSchema(BaseModel):
    start_date: Optional[str] = None  # date string
    end_date: Optional[str] = None  # date string
    service_key: Optional[str] = None
    provider_id: Optional[str] = None


class CustomerBreakdownSchema(BaseModel):
    customer_key: str
    total_cost: float
    event_count: int


class CustomerTokenBreakdownSchema(BaseModel):
    customer_key: str
    total_tokens: float
    event_count: int



from __future__ import annotations

import pytest
from pydantic import ValidationError

from aicostmanager.models import (
    Condition,
    CostUnitSwitch,
    CustomCostUnitIn,
    CustomCostUnitOut,
    CustomServiceConfiguration,
    CustomServiceFilter,
    CustomServiceIn,
    CustomServiceOut,
    CustomServiceSummaryOut,
    Mapping,
    MappingGroup,
    RoundingConfig,
)


class TestCustomCostUnitModels:
    """Test custom cost unit models."""

    def test_custom_cost_unit_in_valid(self):
        """Test valid CustomCostUnitIn creation."""
        unit = CustomCostUnitIn(
            name="api-calls",
            cost="0.05",
            unit="request",
            per_quantity=1,
            min_units=0,
            max_units=10000,
            currency="USD",
            is_active=True,
        )
        assert unit.name == "api-calls"
        assert unit.cost == "0.05"
        assert unit.unit == "request"
        assert unit.currency == "USD"
        assert unit.is_active is True

    def test_custom_cost_unit_in_defaults(self):
        """Test CustomCostUnitIn with default values."""
        unit = CustomCostUnitIn(
            name="tokens",
            cost=0.001,
            unit="token",
        )
        assert unit.per_quantity == 1
        assert unit.min_units == 0
        assert unit.max_units == 10000000
        assert unit.currency == "USD"
        assert unit.is_active is True

    def test_custom_cost_unit_out_valid(self):
        """Test valid CustomCostUnitOut creation."""
        unit = CustomCostUnitOut(
            name="api-calls",
            cost="0.05",
            unit="request",
            per_quantity=1,
            min_units=0,
            max_units=10000,
            currency="USD",
            is_active=True,
        )
        assert unit.name == "api-calls"
        assert unit.cost == "0.05"


class TestMappingModels:
    """Test mapping-related models."""

    def test_rounding_config_valid(self):
        """Test valid RoundingConfig creation."""
        config = RoundingConfig(
            mode="ceil",
            increment=1000,
        )
        assert config.mode == "ceil"
        assert config.increment == 1000

    def test_rounding_config_minimal(self):
        """Test RoundingConfig with minimal fields."""
        config = RoundingConfig(mode="round")
        assert config.mode == "round"
        assert config.increment is None

    def test_condition_equals(self):
        """Test Condition with equals."""
        condition = Condition(
            equals={"path": "tier", "value": "premium"}
        )
        assert condition.equals == {"path": "tier", "value": "premium"}
        assert condition.regex is None

    def test_condition_regex(self):
        """Test Condition with regex."""
        condition = Condition(
            regex={"path": "user_id", "pattern": "^bot_"}
        )
        assert condition.regex == {"path": "user_id", "pattern": "^bot_"}
        assert condition.equals is None

    def test_cost_unit_switch_valid(self):
        """Test valid CostUnitSwitch creation."""
        switch = CostUnitSwitch(
            field="plan",
            cases={"premium": "premium-tier", "enterprise": "enterprise-tier"},
            default="basic-tier",
        )
        assert switch.field == "plan"
        assert switch.cases["premium"] == "premium-tier"
        assert switch.default == "basic-tier"

    def test_mapping_basic(self):
        """Test basic Mapping creation."""
        mapping = Mapping(
            cost_unit="api-calls",
            field="request_count",
        )
        assert mapping.cost_unit == "api-calls"
        assert mapping.field == "request_count"
        assert mapping.divide_by is None
        assert mapping.multiply_by is None

    def test_mapping_advanced(self):
        """Test advanced Mapping with all options."""
        rounding = RoundingConfig(mode="ceil", increment=1000)
        condition = Condition(equals={"path": "model", "value": "gpt-4"})

        mapping = Mapping(
            cost_unit="tokens",
            field="usage.input_tokens",
            divide_by=1000,
            multiply_by=1.1,
            min_value=1,
            max_value=100000,
            rounding=rounding,
            only_if=condition,
        )
        assert mapping.cost_unit == "tokens"
        assert mapping.divide_by == 1000
        assert mapping.multiply_by == 1.1
        assert mapping.min_value == 1
        assert mapping.max_value == 100000
        assert mapping.rounding.mode == "ceil"
        assert mapping.only_if.equals["value"] == "gpt-4"

    def test_mapping_with_cost_unit_switch(self):
        """Test Mapping with cost unit switch."""
        switch = CostUnitSwitch(
            field="tier",
            cases={"premium": "premium-tokens"},
            default="basic-tokens",
        )

        mapping = Mapping(
            cost_unit_switch=switch,
            field="usage.tokens",
            divide_by=1000,
        )
        assert mapping.cost_unit is None
        assert mapping.cost_unit_switch.field == "tier"
        assert mapping.divide_by == 1000

    def test_mapping_group_valid(self):
        """Test valid MappingGroup creation."""
        condition = Condition(equals={"path": "tier", "value": "premium"})
        mapping = Mapping(cost_unit="premium-tokens", field="usage.tokens")

        group = MappingGroup(
            when=condition,
            mappings=[mapping],
        )
        assert group.when.equals["value"] == "premium"
        assert len(group.mappings) == 1
        assert group.mappings[0].cost_unit == "premium-tokens"


class TestCustomServiceConfiguration:
    """Test custom service configuration models."""

    def test_configuration_basic(self):
        """Test basic CustomServiceConfiguration."""
        cost_unit = CustomCostUnitIn(
            name="requests",
            cost="0.01",
            unit="request",
        )
        mapping = Mapping(
            cost_unit="requests",
            field="request_count",
        )

        config = CustomServiceConfiguration(
            cost_units=[cost_unit],
            mappings=[mapping],
        )
        assert len(config.cost_units) == 1
        assert len(config.mappings) == 1
        assert config.mapping_groups is None
        assert config.exclusions is None

    def test_configuration_full(self):
        """Test full CustomServiceConfiguration with all features."""
        # Cost units
        cost_units = [
            CustomCostUnitIn(
                name="basic-tier",
                cost="0.10",
                unit="request",
                currency="USD",
            ),
            CustomCostUnitIn(
                name="premium-tier",
                cost="0.05",
                unit="request",
                currency="USD",
            ),
        ]

        # Mappings
        mappings = [
            Mapping(
                cost_unit_switch=CostUnitSwitch(
                    field="plan",
                    cases={"premium": "premium-tier"},
                    default="basic-tier",
                ),
                field="request_count",
            ),
        ]

        # Mapping groups
        condition = Condition(equals={"path": "environment", "value": "production"})
        mapping_groups = [
            MappingGroup(
                when=condition,
                mappings=[
                    Mapping(cost_unit="production-tokens", field="usage.tokens"),
                ],
            ),
        ]

        # Exclusions
        exclusions = [
            Condition(equals={"path": "environment", "value": "test"}),
            Condition(regex={"path": "user_id", "pattern": "^bot_"}),
        ]

        config = CustomServiceConfiguration(
            cost_units=cost_units,
            mappings=mappings,
            mapping_groups=mapping_groups,
            exclusions=exclusions,
        )
        assert len(config.cost_units) == 2
        assert len(config.mappings) == 1
        assert len(config.mapping_groups) == 1
        assert len(config.exclusions) == 2

    def test_configuration_validation_missing_cost_units(self):
        """Test that configuration requires cost_units."""
        with pytest.raises(ValidationError):
            CustomServiceConfiguration(
                mappings=[Mapping(cost_unit="test", field="count")],
            )


class TestCustomServiceModels:
    """Test custom service input/output models."""

    def test_custom_service_in_valid(self):
        """Test valid CustomServiceIn creation."""
        config = CustomServiceConfiguration(
            cost_units=[
                CustomCostUnitIn(
                    name="requests",
                    cost="0.01",
                    unit="request",
                ),
            ],
            mappings=[
                Mapping(
                    cost_unit="requests",
                    field="request_count",
                ),
            ],
        )

        service = CustomServiceIn(
            custom_service_key="my-api-service",
            configuration=config,
            is_active=True,
        )
        assert service.custom_service_key == "my-api-service"
        assert service.is_active is True
        assert len(service.configuration.cost_units) == 1

    def test_custom_service_in_validation_key_required(self):
        """Test that custom_service_key is required."""
        config = CustomServiceConfiguration(
            cost_units=[CustomCostUnitIn(name="test", cost="1", unit="test")],
        )

        with pytest.raises(ValidationError):
            CustomServiceIn(
                configuration=config,
            )

    def test_custom_service_in_validation_config_required(self):
        """Test that configuration is required."""
        with pytest.raises(ValidationError):
            CustomServiceIn(
                custom_service_key="test-service",
            )

    def test_custom_service_out_valid(self):
        """Test valid CustomServiceOut creation."""
        config = CustomServiceConfiguration(
            cost_units=[
                CustomCostUnitIn(name="requests", cost="0.01", unit="request"),
            ],
        )

        cost_units_out = [
            CustomCostUnitOut(name="requests", cost="0.01", unit="request", per_quantity=1, min_units=0, max_units=10000000, currency="USD", is_active=True),
        ]

        service = CustomServiceOut(
            uuid="service-uuid",
            custom_service_key="my-api-service",
            configuration=config,
            is_active=True,
            is_deleted=False,
            created_at="2024-01-01T00:00:00Z",
            updated_at="2024-01-01T00:00:00Z",
            cost_units=cost_units_out,
            team_uuid="team-uuid",
            team_name="My Team",
        )
        assert service.uuid == "service-uuid"
        assert service.custom_service_key == "my-api-service"
        assert service.team_name == "My Team"

    def test_custom_service_summary_out_valid(self):
        """Test valid CustomServiceSummaryOut creation."""
        summary = CustomServiceSummaryOut(
            uuid="service-uuid",
            custom_service_key="my-api-service",
            is_active=True,
            is_deleted=False,
            created_at="2024-01-01T00:00:00Z",
            updated_at="2024-01-01T00:00:00Z",
            cost_units_count=2,
        )
        assert summary.uuid == "service-uuid"
        assert summary.cost_units_count == 2


class TestCustomServiceFilter:
    """Test custom service filter model."""

    def test_filter_empty(self):
        """Test empty filter."""
        filter_obj = CustomServiceFilter()
        assert filter_obj.is_active is None
        assert filter_obj.is_deleted is None
        assert filter_obj.has_cost_units is None

    def test_filter_with_values(self):
        """Test filter with values."""
        filter_obj = CustomServiceFilter(
            is_active=True,
            is_deleted=False,
            has_cost_units=True,
        )
        assert filter_obj.is_active is True
        assert filter_obj.is_deleted is False
        assert filter_obj.has_cost_units is True


class TestComplexExamples:
    """Test complex real-world examples from documentation."""

    def test_simple_api_pricing_example(self):
        """Test the simple API pricing example from docs."""
        config = CustomServiceConfiguration(
            cost_units=[
                CustomCostUnitIn(
                    name="requests",
                    cost="0.01",
                    unit="request",
                    per_quantity=1,
                    currency="USD",
                    is_active=True,
                ),
            ],
            mappings=[
                Mapping(
                    cost_unit="requests",
                    field="request_count",
                ),
            ],
        )

        service = CustomServiceIn(
            custom_service_key="my-api-service",
            configuration=config,
            is_active=True,
        )
        assert service.custom_service_key == "my-api-service"

    def test_token_based_llm_pricing_example(self):
        """Test the token-based LLM pricing example from docs."""
        config = CustomServiceConfiguration(
            cost_units=[
                CustomCostUnitIn(
                    name="input-tokens",
                    cost="0.001",
                    unit="token",
                    per_quantity=1000,
                    currency="USD",
                    is_active=True,
                ),
                CustomCostUnitIn(
                    name="output-tokens",
                    cost="0.002",
                    unit="token",
                    per_quantity=1000,
                    currency="USD",
                    is_active=True,
                ),
            ],
            mappings=[
                Mapping(
                    cost_unit="input-tokens",
                    field="usage.prompt_tokens",
                    divide_by=1000,
                ),
                Mapping(
                    cost_unit="output-tokens",
                    field="usage.completion_tokens",
                    divide_by=1000,
                ),
            ],
        )

        service = CustomServiceIn(
            custom_service_key="custom-llm",
            configuration=config,
            is_active=True,
        )
        assert len(service.configuration.cost_units) == 2
        assert len(service.configuration.mappings) == 2

    def test_complex_pricing_with_tiers_example(self):
        """Test the complex pricing with tiers example from docs."""
        config = CustomServiceConfiguration(
            cost_units=[
                CustomCostUnitIn(
                    name="basic-tier",
                    cost="0.10",
                    unit="request",
                    per_quantity=1,
                    currency="USD",
                    is_active=True,
                ),
                CustomCostUnitIn(
                    name="premium-tier",
                    cost="0.05",
                    unit="request",
                    per_quantity=1,
                    currency="USD",
                    is_active=True,
                ),
            ],
            mappings=[
                Mapping(
                    cost_unit_switch=CostUnitSwitch(
                        field="plan",
                        cases={"premium": "premium-tier"},
                        default="basic-tier",
                    ),
                    field="request_count",
                ),
            ],
        )

        service = CustomServiceIn(
            custom_service_key="tiered-api",
            configuration=config,
            is_active=True,
        )
        assert len(service.configuration.cost_units) == 2
        assert service.configuration.mappings[0].cost_unit_switch is not None



import json
import time
import urllib.request
import uuid

import pytest

boto3 = pytest.importorskip("boto3")

from aicostmanager.delivery import DeliveryConfig, DeliveryType, create_delivery
from aicostmanager.ini_manager import IniManager
from aicostmanager.tracker import Tracker
from aicostmanager.usage_utils import get_usage_from_response

BASE_URL = "http://127.0.0.1:8001"


def _wait_for_cost_event(aicm_api_key: str, response_id: str, timeout: int = 30):
    headers = {"Authorization": f"Bearer {aicm_api_key}"}
    deadline = time.time() + timeout
    while time.time() < deadline:
        try:
            req = urllib.request.Request(
                f"{BASE_URL}/api/v1/cost-events/{response_id}",
                headers=headers,
            )
            with urllib.request.urlopen(req, timeout=5) as resp:
                if resp.status == 200:
                    data = json.load(resp)
                    # Support both array and object responses
                    if isinstance(data, list) and data:
                        first = data[0]
                        event_id = first.get("event_id") or first.get("uuid")
                        if event_id:
                            uuid.UUID(str(event_id))
                            return data
                    elif isinstance(data, dict):
                        event_id = data.get("event_id") or data.get(
                            "cost_event", {}
                        ).get("event_id")
                        if event_id:
                            uuid.UUID(str(event_id))
                            return data
        except Exception:
            pass
        time.sleep(1)
    raise AssertionError(f"cost event for {response_id} not found")


@pytest.mark.parametrize(
    "service_key, model",
    [
        (
            "amazon-bedrock::us.meta.llama3-3-70b-instruct-v1:0",
            "us.meta.llama3-3-70b-instruct-v1:0",
        ),
        ("amazon-bedrock::us.amazon.nova-pro-v1:0", "us.amazon.nova-pro-v1:0"),
    ],
)
def test_bedrock_tracker(service_key, model, aws_region, aicm_api_key):
    if not aws_region:
        pytest.skip("AWS_DEFAULT_REGION not set in .env file")
    ini = IniManager("ini")
    dconfig = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=aicm_api_key,
        aicm_api_base=BASE_URL,
    )
    delivery = create_delivery(
        DeliveryType.PERSISTENT_QUEUE, dconfig, poll_interval=0.1, batch_interval=0.1
    )
    tracker = Tracker(
        aicm_api_key=aicm_api_key, ini_path=ini.ini_path, delivery=delivery
    )
    client = boto3.client("bedrock-runtime", region_name=aws_region)

    body = {
        "messages": [{"role": "user", "content": [{"text": "Say hi"}]}],
        "inferenceConfig": {"maxTokens": 50},
    }
    # Background tracking via queue
    try:
        resp = client.converse(modelId=model, **body)
    except Exception as e:
        msg = str(e)
        if (
            "on-demand throughput isn't supported" in msg
            or "on-demand throughput isn’nt supported" in msg
        ):
            pytest.skip(
                "Bedrock model requires provisioned throughput; skipping this case"
            )
        raise
    response_id = resp.get("ResponseMetadata", {}).get("RequestId") or resp.get(
        "output", {}
    ).get("message", {}).get("id")
    usage_payload = get_usage_from_response(resp, "amazon-bedrock")
    tracker.track(service_key, usage_payload, response_id=response_id)
    _wait_for_cost_event(aicm_api_key, response_id)

    # Immediate delivery
    body2 = {
        "messages": [{"role": "user", "content": [{"text": "Say hi again"}]}],
        "inferenceConfig": {"maxTokens": 50},
    }
    try:
        resp2 = client.converse(modelId=model, **body2)
    except Exception as e:
        msg = str(e)
        if (
            "on-demand throughput isn't supported" in msg
            or "on-demand throughput isn’nt supported" in msg
        ):
            pytest.skip(
                "Bedrock model requires provisioned throughput; skipping this case"
            )
        raise
    response_id2 = resp2.get("ResponseMetadata", {}).get("RequestId") or resp2.get(
        "output", {}
    ).get("message", {}).get("id")
    dconfig2 = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=aicm_api_key,
        aicm_api_base=BASE_URL,
    )
    delivery2 = create_delivery(DeliveryType.IMMEDIATE, dconfig2)
    with Tracker(
        aicm_api_key=aicm_api_key, ini_path=ini.ini_path, delivery=delivery2
    ) as t2:
        usage2 = get_usage_from_response(resp2, "amazon-bedrock")
        t2.track(service_key, usage2, response_id=response_id2)
    _wait_for_cost_event(aicm_api_key, response_id2)

    tracker.close()



import pathlib
import time

import jwt

from aicostmanager.client import CostManagerClient
from aicostmanager.config_manager import ConfigManager
from aicostmanager.limits import TriggeredLimitManager, UsageLimitManager
from aicostmanager.models import (
    Period,
    ThresholdType,
    UsageLimitIn,
    UsageLimitOut,
    UsageLimitProgressOut,
)

PRIVATE_KEY = (pathlib.Path(__file__).parent / "threshold_private_key.pem").read_text()
PUBLIC_KEY = (pathlib.Path(__file__).parent / "threshold_public_key.pem").read_text()


def _make_triggered_limits():
    now = int(time.time())
    event = {
        "event_id": "evt-api-key-limit",
        "limit_id": "lmt-api-key-limit",
        "threshold_type": "limit",
        "amount": 100.0,
        "period": "month",
        "limit_context": "key",
        "limit_message": "Usage limit exceeded",
        "service_key": "openai::gpt-4",
        "customer_key": "api-key-customer",
        "api_key_id": "550e8400-e29b-41d4-a716-446655440000",
        "triggered_at": "2024-12-31T18:00:00Z",
        "expires_at": "2025-01-01T18:00:00Z",
    }
    payload = {
        "iss": "aicm-api",
        "sub": "550e8400-e29b-41d4-a716-446655440000",
        "iat": now,
        "jti": "tl",
        "version": "v1",
        "key_id": "test",
        "triggered_limits": [event],
    }
    token = jwt.encode(payload, PRIVATE_KEY, algorithm="RS256", headers={"kid": "test"})
    item = {
        "version": "v1",
        "public_key": PUBLIC_KEY,
        "key_id": "test",
        "encrypted_payload": token,
    }
    return item, event


def test_update_and_check(monkeypatch, tmp_path):
    ini = tmp_path / "AICM.ini"
    client = CostManagerClient(aicm_api_key="sk-test", aicm_ini_path=str(ini))
    cfg_mgr = ConfigManager(client)
    tl_mgr = TriggeredLimitManager(client, cfg_mgr)

    item, event = _make_triggered_limits()
    monkeypatch.setattr(client, "get_triggered_limits", lambda: item)

    tl_mgr.update_triggered_limits()

    stored = cfg_mgr.read_triggered_limits()
    assert stored.get("encrypted_payload") == item["encrypted_payload"]

    cfg_matches = cfg_mgr.get_triggered_limits(service_key=event["service_key"])
    assert len(cfg_matches) == 1

    matches = tl_mgr.check_triggered_limits(
        api_key_id=event["api_key_id"], service_key=event["service_key"]
    )
    assert len(matches) == 1
    assert matches[0].limit_id == event["limit_id"]

    no_match = tl_mgr.check_triggered_limits(
        api_key_id=event["api_key_id"], service_key="other-service"
    )
    assert no_match == []

    wrong_api = tl_mgr.check_triggered_limits(api_key_id="different")
    assert wrong_api == []


def test_usage_limit_management(monkeypatch, tmp_path):
    ini = tmp_path / "AICM.ini"
    client = CostManagerClient(aicm_api_key="sk-test", aicm_ini_path=str(ini))
    ul_mgr = UsageLimitManager(client)

    limit = UsageLimitOut(
        uuid="lim1",
        threshold_type=ThresholdType.LIMIT,
        amount=1,
        period=Period.DAY,
        service=None,
        client=None,
        notification_list=None,
        active=True,
    )

    progress = UsageLimitProgressOut(
        **limit.model_dump(),
        current_spend=0,
        remaining_amount=1,
    )

    monkeypatch.setattr(client, "list_usage_limits", lambda: [limit])
    assert ul_mgr.list_usage_limits()[0].uuid == "lim1"

    new_limit = UsageLimitIn(
        threshold_type=ThresholdType.LIMIT,
        amount=1,
        period=Period.DAY,
        team_uuid="team1",
    )
    monkeypatch.setattr(client, "create_usage_limit", lambda data: limit)
    assert ul_mgr.create_usage_limit(new_limit).uuid == "lim1"

    monkeypatch.setattr(client, "get_usage_limit", lambda lid: limit)
    assert ul_mgr.get_usage_limit("lim1").uuid == "lim1"

    updated = limit.model_copy(update={"amount": 2})
    monkeypatch.setattr(client, "update_usage_limit", lambda lid, data: updated)
    assert ul_mgr.update_usage_limit("lim1", new_limit).amount == 2

    called = {}

    def _delete(lid):
        called["id"] = lid

    monkeypatch.setattr(client, "delete_usage_limit", _delete)
    ul_mgr.delete_usage_limit("lim1")
    assert called["id"] == "lim1"

    monkeypatch.setattr(client, "list_usage_limit_progress", lambda: [progress])
    assert ul_mgr.list_usage_limit_progress()[0].current_spend == 0



from __future__ import annotations

import logging
import os
import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

import httpx
from tenacity import (
    Retrying,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential_jitter,
)

from ..client.exceptions import UsageLimitExceeded
from ..config_manager import ConfigManager
from ..ini_manager import IniManager
from ..logger import create_logger


class DeliveryType(str, Enum):
    IMMEDIATE = "immediate"
    PERSISTENT_QUEUE = "persistent_queue"


@dataclass
class DeliveryConfig:
    """Common configuration shared by delivery implementations."""

    ini_manager: IniManager
    aicm_api_key: str | None = None
    aicm_api_base: str | None = None
    aicm_api_url: str | None = None
    timeout: float = 10.0
    transport: httpx.BaseTransport | None = None
    log_file: str | None = None
    log_level: str | None = None
    # For immediate delivery post-send wait before checking limits
    immediate_pause_seconds: float = 5.0


class Delivery(ABC):
    """Abstract base class for tracker delivery mechanisms."""

    def __init__(
        self,
        config: DeliveryConfig,
        *,
        endpoint: str = "/track",
        body_key: str = "tracked",
        logger: logging.Logger | None = None,
    ) -> None:
        if config.ini_manager is None:
            raise ValueError("ini_manager must be provided")
        self.ini_manager = config.ini_manager
        self.logger = logger or create_logger(
            self.__class__.__name__, config.log_file, config.log_level
        )
        self.api_key = config.aicm_api_key or os.getenv("AICM_API_KEY")
        self.api_base = config.aicm_api_base or "https://aicostmanager.com"
        self.api_url = config.aicm_api_url or "/api/v1"
        self.timeout = config.timeout
        self._transport = config.transport
        self._client = httpx.Client(timeout=config.timeout, transport=config.transport)
        self._root = self.api_base.rstrip("/") + self.api_url.rstrip("/")
        self._endpoint = self._root + endpoint
        self._body_key = body_key
        self._headers = {
            "Authorization": f"Bearer {self.api_key}",
            "User-Agent": "aicostmanager-python",
        }
        self.immediate_pause_seconds = getattr(config, "immediate_pause_seconds", 5.0)

    def _limits_enabled(self) -> bool:
        """Return ``True`` when triggered limits should be enforced."""
        val = self.ini_manager.get_option("tracker", "AICM_LIMITS_ENABLED", "false")
        return str(val).lower() in {"1", "true", "yes", "on"}

    def _post_with_retry(
        self, body: Dict[str, Any], *, max_attempts: int
    ) -> Dict[str, Any]:
        def _retryable(exc: Exception) -> bool:
            if isinstance(exc, httpx.HTTPStatusError):
                return exc.response is None or exc.response.status_code >= 500
            # Don't retry if client is closed - it won't recover
            if isinstance(exc, RuntimeError) and "client has been closed" in str(exc):
                return False
            return True

        for attempt in Retrying(
            stop=stop_after_attempt(max_attempts),
            wait=wait_exponential_jitter(),
            retry=retry_if_exception(_retryable),
            reraise=True,
        ):
            with attempt:
                # Check if client is closed before attempting request
                if self._client.is_closed:
                    raise RuntimeError(
                        "HTTP client has been closed - cannot send request"
                    )
                resp = self._client.post(
                    self._endpoint, json=body, headers=self._headers
                )
                resp.raise_for_status()
                return resp.json()
        raise RuntimeError("unreachable")

    def _check_triggered_limits(self, payload: Dict[str, Any]) -> None:
        """Raise ``UsageLimitExceeded`` if ``payload`` matches a triggered limit.

        Uses a client-backed ConfigManager to ensure we can refresh and decode
        triggered limits when the server omits the public key from inline responses.
        """
        # Use file-backed ConfigManager to avoid incidental GETs that could stale/clear state
        cfg = ConfigManager(ini_path=self.ini_manager.ini_path, load=False)
        service_key = payload.get("service_key")
        customer_key = payload.get("customer_key")
        limits = cfg.get_triggered_limits(
            service_key=service_key,
            customer_key=customer_key,
        )
        # Extract the API key ID (UUID suffix after the last dot)
        api_key_id = None
        if self.api_key and "." in self.api_key:
            api_key_id = self.api_key.split(".")[-1]
        if api_key_id:
            limits = [l for l in limits if l.api_key_id == api_key_id]
        if limits:
            # Recompute with latest state (limits may have been updated by recent track)
            limits = cfg.get_triggered_limits(
                service_key=service_key,
                customer_key=customer_key,
            )
            if api_key_id:
                limits = [l for l in limits if l.api_key_id == api_key_id]
            if limits:
                raise UsageLimitExceeded(limits)

    @abstractmethod
    def _enqueue(self, payload: Dict[str, Any]) -> Any:
        """Implementation-specific enqueue logic."""

    def enqueue(self, payload: Dict[str, Any]) -> Any:
        """Queue ``payload`` for background delivery and enforce triggered limits."""
        if isinstance(self, QueueDelivery):
            result = self._enqueue(payload)
            if self._limits_enabled():
                self._check_triggered_limits(
                    payload
                )  # This check happens AFTER enqueue
            return result

        result = self._enqueue(payload)
        if self._limits_enabled():
            self._check_triggered_limits(payload)
        return result

    def deliver(self, body: Dict[str, Any]) -> None:
        """Queue payloads from a pre-built request body."""
        for payload in body.get(self._body_key, []):
            self.enqueue(payload)

    def stop(self) -> None:  # pragma: no cover - default no-op
        """Shutdown any background resources."""
        return None


@dataclass
class QueueItem:
    payload: Dict[str, Any]
    id: Optional[int] = None
    retry_count: int = 0


class QueueWorker(ABC):
    """Abstract helper for queue based deliveries."""

    @abstractmethod
    def get_batch(self, max_batch_size: int, *, block: bool = True) -> List[QueueItem]:
        """Return up to ``max_batch_size`` items for processing."""

    def acknowledge(
        self, items: List[QueueItem]
    ) -> None:  # pragma: no cover - default no-op
        return None

    def reschedule(self, item: QueueItem) -> None:  # pragma: no cover - default no-op
        return None

    def queued(self) -> int:  # pragma: no cover - default no-op
        return 0


class QueueDelivery(Delivery, QueueWorker):
    """Base class for threaded queue based deliveries."""

    def __init__(
        self,
        config: DeliveryConfig,
        *,
        batch_interval: float = 0.5,
        max_batch_size: int = 1000,
        max_attempts: int = 5,
        max_retries: int = 5,
        **kwargs: Any,
    ) -> None:
        super().__init__(config, **kwargs)
        self.batch_interval = batch_interval
        self.max_batch_size = max_batch_size
        self.max_attempts = max_attempts
        self.max_retries = max_retries
        self._total_sent = 0
        self._total_failed = 0
        self._stop = threading.Event()
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()

    def _process_batch(self, batch: List[QueueItem]) -> None:
        payloads = [item.payload for item in batch]
        body = {self._body_key: payloads}
        try:
            data = self._post_with_retry(body, max_attempts=self.max_attempts)
            if self._limits_enabled() and isinstance(data, dict):
                tl_data = data.get("triggered_limits")
                if tl_data:
                    # Use a ConfigManager that preserves existing INI settings
                    cfg = ConfigManager(ini_path=self.ini_manager.ini_path, load=True)
                    try:
                        cfg.write_triggered_limits(tl_data)
                    except Exception as exc:  # pragma: no cover - network failures
                        self.logger.error("Triggered limits update failed: %s", exc)
        except Exception as exc:  # pragma: no cover - network failures
            self.logger.error("Delivery failed: %s", exc)
            for item in batch:
                item.retry_count += 1
                self.reschedule(item)
                self._total_failed += 1
        else:
            self.acknowledge(batch)
            self._total_sent += len(batch)

    def _run(self) -> None:
        while not self._stop.is_set():
            batch = self.get_batch(self.max_batch_size, block=True)
            if not batch:
                continue
            self._process_batch(batch)
        while True:
            batch = self.get_batch(self.max_batch_size, block=False)
            if not batch:
                break
            self._process_batch(batch)
        self._client.close()

    def stop(self) -> None:
        self._stop.set()
        self._thread.join()
        super().stop()

    def stats(self) -> Dict[str, Any]:
        return {
            "queued": self.queued(),
            "total_sent": self._total_sent,
            "total_failed": self._total_failed,
        }



import os
import time

import httpx
import pytest

from aicostmanager import Tracker
from aicostmanager.delivery import DeliveryConfig, DeliveryType, create_delivery
from aicostmanager.ini_manager import IniManager
from tests.track_asserts import assert_track_result_payload

VALID_PAYLOAD = {
    "prompt_tokens": 19,
    "completion_tokens": 10,
    "total_tokens": 29,
    "prompt_tokens_details": {
        "cached_tokens": 0,
        "audio_tokens": 0,
    },
    "completion_tokens_details": {
        "reasoning_tokens": 0,
        "audio_tokens": 0,
        "accepted_prediction_tokens": 0,
        "rejected_prediction_tokens": 0,
    },
}


def _make_tracker(api_key: str, api_base: str, tmp_path) -> Tracker:
    ini = IniManager(str(tmp_path / "ini"))
    dconfig = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=api_key,
        aicm_api_base=api_base,
    )
    delivery = create_delivery(
        DeliveryType.PERSISTENT_QUEUE,
        dconfig,
        db_path=str(tmp_path / "queue.db"),
        poll_interval=0.1,
        batch_interval=0.1,
        max_attempts=2,  # Reduce attempts for faster failure
        max_retries=2,  # Reduce retries for faster failure
    )
    return Tracker(aicm_api_key=api_key, ini_path=ini.ini_path, delivery=delivery)


def _wait_for_empty(delivery, timeout: float = 10.0) -> bool:
    for _ in range(int(timeout / 0.1)):
        stats = getattr(delivery, "stats", lambda: {})()
        if stats.get("queued", 0) == 0:
            return True
        time.sleep(0.1)
    return False


def test_track_single_event_success(aicm_api_key, aicm_api_base, tmp_path):
    tracker = _make_tracker(aicm_api_key, aicm_api_base, tmp_path)
    tracker.track(
        "openai_chat",
        VALID_PAYLOAD,
        response_id="evt1",
        timestamp="2025-01-01T00:00:00Z",
    )
    assert _wait_for_empty(tracker.delivery)
    tracker.close()


def test_track_multiple_events_with_errors(aicm_api_key, aicm_api_base, tmp_path):
    tracker = _make_tracker(aicm_api_key, aicm_api_base, tmp_path)
    events = [
        {
            "api_id": "openai_chat",
            "service_key": "openai::gpt-5-mini",
            "response_id": "ok1",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Missing service_key
            "api_id": "openai_chat",
            "response_id": "missing",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Invalid service_key format
            "api_id": "openai_chat",
            "service_key": "invalidformat",
            "response_id": "badformat",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Service not found
            "api_id": "openai_chat",
            "service_key": "openai::does-not-exist",
            "response_id": "noservice",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # API client not found
            "api_id": "nonexistent_client",
            "service_key": "openai::gpt-5-mini",
            "response_id": "noapi",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Payload validation error (missing total_tokens)
            "api_id": "openai_chat",
            "service_key": "openai::gpt-5-mini",
            "response_id": "badpayload",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": {
                "prompt_tokens": 19,
                "completion_tokens": 10,
                "prompt_tokens_details": {
                    "cached_tokens": 0,
                    "audio_tokens": 0,
                },
                "completion_tokens_details": {
                    "reasoning_tokens": 0,
                    "audio_tokens": 0,
                    "accepted_prediction_tokens": 0,
                    "rejected_prediction_tokens": 0,
                },
            },
        },
    ]

    for event in events:
        tracker.track(
            event["api_id"],
            event["payload"],
            response_id=event.get("response_id"),
            timestamp=event.get("timestamp"),
        )

    assert _wait_for_empty(tracker.delivery)
    tracker.close()


def test_deliver_now_single_event_success(aicm_api_key, aicm_api_base):
    ini = IniManager("ini")
    # Disable limits for this test since it's not testing limits functionality
    ini.set_option("tracker", "AICM_LIMITS_ENABLED", "false")
    dconfig = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=aicm_api_key,
        aicm_api_base=aicm_api_base,
    )
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig)
    with Tracker(
        aicm_api_key=aicm_api_key, ini_path="ini", delivery=delivery
    ) as tracker:
        result = tracker.track(
            "openai::gpt-5-mini",
            VALID_PAYLOAD,
            response_id="evt1",
            timestamp="2025-01-01T00:00:00Z",
        )
        if result is None:
            pytest.skip(
                "Server rejected tracking request - check server logs for validation errors"
            )
        payload = result.get("result") or {}
        assert payload, "Immediate track should return a result payload"
        assert_track_result_payload(payload)
        assert "triggered_limits" in result


@pytest.mark.skipif(
    os.environ.get("RUN_NETWORK_TESTS") != "1", reason="requires network access"
)
def test_deliver_now_multiple_events_with_errors(aicm_api_key, aicm_api_base):
    ini = IniManager("ini")
    # Disable limits for this test since it's not testing limits functionality
    ini.set_option("tracker", "AICM_LIMITS_ENABLED", "false")
    dconfig = DeliveryConfig(
        ini_manager=ini,
        aicm_api_key=aicm_api_key,
        aicm_api_base=aicm_api_base,
    )
    delivery = create_delivery(DeliveryType.IMMEDIATE, dconfig)
    tracker = Tracker(aicm_api_key=aicm_api_key, ini_path="ini", delivery=delivery)
    events = [
        {
            "api_id": "openai_chat",
            "service_key": "openai::gpt-5-mini",
            "response_id": "ok1",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Missing service_key
            "api_id": "openai_chat",
            "response_id": "missing",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Invalid service_key format
            "api_id": "openai_chat",
            "service_key": "invalidformat",
            "response_id": "badformat",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Service not found
            "api_id": "openai_chat",
            "service_key": "openai::does-not-exist",
            "response_id": "noservice",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # API client not found
            "api_id": "nonexistent_client",
            "service_key": "openai::gpt-5-mini",
            "response_id": "noapi",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": VALID_PAYLOAD,
        },
        {
            # Payload validation error (missing total_tokens)
            "api_id": "openai_chat",
            "service_key": "openai::gpt-5-mini",
            "response_id": "badpayload",
            "timestamp": "2025-01-01T00:00:00Z",
            "payload": {
                "prompt_tokens": 19,
                "completion_tokens": 10,
                "prompt_tokens_details": {
                    "cached_tokens": 0,
                    "audio_tokens": 0,
                },
                "completion_tokens_details": {
                    "reasoning_tokens": 0,
                    "audio_tokens": 0,
                    "accepted_prediction_tokens": 0,
                    "rejected_prediction_tokens": 0,
                },
            },
        },
    ]

    def _extract_payload(data: dict, fallback_id: str) -> dict:
        results = data.get("results", [])
        if results and isinstance(results[0], dict):
            payload = results[0]
            payload.setdefault("response_id", fallback_id)
            return payload
        errors = data.get("errors") or data.get("detail")
        if isinstance(errors, str):
            errors = [errors]
        elif not isinstance(errors, list):
            errors = ["Rejected by server"]
        return {"response_id": fallback_id, "errors": errors}

    collected: dict[str, dict] = {}

    for idx, event in enumerate(events):
        response_id = event.get("response_id") or f"evt-{idx}"
        if idx == 1:
            body = {
                "tracked": [
                    {
                        "api_id": event["api_id"],
                        "response_id": response_id,
                        "timestamp": event["timestamp"],
                        "payload": event["payload"],
                    }
                ]
            }
            with httpx.Client() as client:
                resp = client.post(
                    f"{aicm_api_base.rstrip('/')}/api/v1/track",
                    json=body,
                    headers={"Authorization": f"Bearer {aicm_api_key}"},
                )
            data = resp.json()
            collected_response = _extract_payload(data, response_id)
            collected[collected_response["response_id"]] = collected_response
            continue

        if idx in (2, 3, 5):
            body = {
                "tracked": [
                    {
                        "api_id": event["api_id"],
                        "service_key": event.get("service_key"),
                        "response_id": response_id,
                        "timestamp": event.get("timestamp"),
                        "payload": event.get("payload"),
                    }
                ]
            }
            with httpx.Client() as client:
                resp = client.post(
                    f"{aicm_api_base.rstrip('/')}/api/v1/track",
                    json=body,
                    headers={"Authorization": f"Bearer {aicm_api_key}"},
                )
            assert resp.status_code in {202, 422}, resp.text
            data = resp.json()
            collected_response = _extract_payload(data, response_id)
            collected[collected_response["response_id"]] = collected_response
            continue

        track_res = tracker.track(
            event["api_id"],
            event["payload"],
            response_id=response_id,
            timestamp=event.get("timestamp"),
        )
        payload = track_res.get("result") if isinstance(track_res, dict) else None
        if isinstance(payload, dict):
            collected[payload.get("response_id", response_id)] = payload
        else:
            collected[response_id] = {
                "response_id": response_id,
                "errors": ["Rejected by server"],
            }

    tracker.close()

    ok_payload = collected.get("ok1")
    assert ok_payload is not None
    assert_track_result_payload(ok_payload)

    missing_payload = collected.get("missing")
    assert missing_payload is not None
    assert missing_payload.get("errors")
    # Accept either specific service_key error or generic rejection
    errors = missing_payload.get("errors", [])
    assert any("service_key" in e.lower() or "rejected" in e.lower() for e in errors)

    badformat_payload = collected.get("badformat")
    assert badformat_payload is not None
    assert badformat_payload.get("errors")

    noservice_payload = collected.get("noservice")
    assert noservice_payload is not None
    # Server may return error response instead of service_key_unknown status
    if noservice_payload.get("errors"):
        assert any(
            "service" in e.lower() or "rejected" in e.lower()
            for e in noservice_payload.get("errors", [])
        )
    else:
        assert_track_result_payload(noservice_payload)
        assert noservice_payload.get("status") == "service_key_unknown"

    noapi_payload = collected.get("noapi")
    if noapi_payload is not None:
        # Some servers may reject this at ingestion time if the API client is unknown
        if noapi_payload.get("errors"):
            # Just verify there are errors - don't check specific content
            assert len(noapi_payload["errors"]) > 0
        else:
            assert_track_result_payload(noapi_payload)

    badpayload_payload = collected.get("badpayload")
    assert badpayload_payload is not None
    assert badpayload_payload.get("errors")


def test_track_missing_response_id_generates_unknown_and_422(
    aicm_api_key, aicm_api_base
):
    # Send directly with httpx to omit response_id entirely.
    with httpx.Client() as client:
        body = {
            "tracked": [
                {
                    "api_id": "openai_chat",
                    "service_key": "openai::gpt-5-mini",
                    "timestamp": "2025-01-01T00:00:00Z",
                    "payload": VALID_PAYLOAD,
                }
            ]
        }
        resp = client.post(
            f"{aicm_api_base.rstrip('/')}/api/v1/track",
            json=body,
            headers={"Authorization": f"Bearer {aicm_api_key}"},
        )
    assert resp.status_code == 422, resp.text
    data = resp.json()
    # New schema returns errors in results array
    results = data.get("results", [])
    if len(results) == 1:
        first = results[0]
        assert "response_id" in first and isinstance(first.get("response_id"), str)
        assert first["response_id"].startswith("UNKNOWN-RESPONSE-")
        assert first.get("errors") == ["Missing response_id."]
    else:
        # Server may not return results for missing response_id
        assert len(results) == 0



from __future__ import annotations

from typing import Any, Dict, Optional

from pydantic import BaseModel, ConfigDict


class GeneratedReportOut(BaseModel):
    report_id: str
    status: str
    created: str  # datetime string
    last_updated: str  # datetime string
    expires_at: str  # datetime string
    parameters: Dict[str, Any]
    download_url: Optional[str] = None

    model_config = ConfigDict(from_attributes=True)



from __future__ import annotations

import os
from typing import Optional

from .exceptions import MissingConfiguration
from ..ini_manager import IniManager


class BaseClient:
    """Shared initialization logic for SDK clients.

    The ``api_base`` and ``api_url`` values are resolved in the following order:

    1. Explicit ``aicm_api_base``/``aicm_api_url`` parameters
    2. Values from ``AICM.INI`` (``tracker`` section)
    3. Environment variables ``AICM_API_BASE``/``AICM_API_URL``
    4. Defaults ``https://aicostmanager.com`` and ``/api/v1``
    """

    def __init__(
        self,
        *,
        aicm_api_key: Optional[str] = None,
        aicm_api_base: Optional[str] = None,
        aicm_api_url: Optional[str] = None,
        aicm_ini_path: Optional[str] = None,
    ) -> None:
        self.ini_manager = IniManager(IniManager.resolve_path(aicm_ini_path))
        self.ini_path = self.ini_manager.ini_path

        self.api_key = aicm_api_key or os.getenv("AICM_API_KEY")

        def _get(option: str, default: str | None = None) -> str | None:
            val = self.ini_manager.get_option("tracker", option)
            if val is not None:
                return val
            return os.getenv(option, default)

        self.api_base = aicm_api_base or _get(
            "AICM_API_BASE", "https://aicostmanager.com"
        )
        self.api_url = aicm_api_url or _get("AICM_API_URL", "/api/v1")
        if not self.api_key:
            raise MissingConfiguration(
                "API key not provided. Set AICM_API_KEY environment variable or pass aicm_api_key"
            )

    @property
    def api_root(self) -> str:
        """Return the combined AICostManager API base URL."""
        return self.api_base.rstrip("/") + self.api_url

    def _store_triggered_limits(self, triggered_limits_response) -> None:
        """Persist triggered limits using the configuration manager."""
        from ..config_manager import ConfigManager

        cfg_mgr = ConfigManager(ini_path=self.ini_path)
        if isinstance(triggered_limits_response, dict):
            tl_data = triggered_limits_response.get(
                "triggered_limits", triggered_limits_response
            )
        else:
            tl_data = triggered_limits_response
        cfg_mgr.write_triggered_limits(tl_data)



from __future__ import annotations

import asyncio
import os
from collections.abc import Iterable, Mapping
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple
from uuid import uuid4

from .client.exceptions import BatchSizeLimitExceeded
from .delivery import (
    Delivery,
    DeliveryConfig,
    DeliveryType,
    create_delivery,
)
from .ini_manager import IniManager
from .logger import create_logger
from .usage_utils import (
    get_streaming_usage_from_response,
    get_usage_from_response,
)


class Tracker:
    """Lightweight usage tracker for the new ``/track`` endpoint."""

    def __init__(
        self,
        *,
        aicm_api_key: str | None = None,
        ini_path: str | None = None,
        delivery: Delivery | None = None,
        delivery_type: DeliveryType | str | None = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> None:
        self.ini_manager = IniManager(IniManager.resolve_path(ini_path))
        self.ini_path = self.ini_manager.ini_path
        self.aicm_api_key = aicm_api_key or os.getenv("AICM_API_KEY")
        ini_dir = Path(self.ini_path).resolve().parent

        def _get(option: str, default: str | None = None) -> str | None:
            val = self.ini_manager.get_option("tracker", option)
            if val is not None:
                return val
            return os.getenv(option, default)

        api_base = _get("AICM_API_BASE", "https://aicostmanager.com")
        api_url = _get("AICM_API_URL", "/api/v1")
        log_file = _get("AICM_LOG_FILE", str(ini_dir / "aicm.log"))
        log_level = _get("AICM_LOG_LEVEL")
        timeout = float(_get("AICM_TIMEOUT", "10.0"))
        poll_interval = float(_get("AICM_POLL_INTERVAL", "0.1"))
        batch_interval = float(_get("AICM_BATCH_INTERVAL", "0.5"))
        immediate_pause_seconds = float(_get("AICM_IMMEDIATE_PAUSE_SECONDS", "5.0"))
        max_attempts = int(_get("AICM_MAX_ATTEMPTS", "3"))
        max_retries = int(_get("AICM_MAX_RETRIES", "5"))
        max_batch_size = int(_get("AICM_MAX_BATCH_SIZE", "1000"))
        log_bodies_val = _get("AICM_LOG_BODIES", "false")
        log_bodies = str(log_bodies_val).lower() in {"1", "true", "yes", "on"}

        raise_on_error_val = _get("AICM_RAISE_ON_ERROR", "false")
        raise_on_error = str(raise_on_error_val).lower() in {"1", "true", "yes", "on"}

        db_path = _get("AICM_DB_PATH", str(ini_dir / "queue.db"))
        delivery_name_cfg = _get("AICM_DELIVERY_TYPE")

        self.logger = create_logger(__name__, log_file, log_level)

        if delivery is not None:
            self.delivery = delivery
            resolved_type = getattr(delivery, "type", None)
        else:
            delivery_name_arg = (
                delivery_type.value
                if isinstance(delivery_type, DeliveryType)
                else delivery_type
            )
            if delivery_name_arg is not None:
                resolved_type = DeliveryType(str(delivery_name_arg).lower())
            elif delivery_name_cfg:
                resolved_type = DeliveryType(delivery_name_cfg.lower())
            else:
                resolved_type = DeliveryType.IMMEDIATE

            final_db_path = (
                db_path if resolved_type == DeliveryType.PERSISTENT_QUEUE else None
            )

            dconfig = DeliveryConfig(
                ini_manager=self.ini_manager,
                aicm_api_key=self.aicm_api_key,
                aicm_api_base=api_base,
                aicm_api_url=api_url,
                timeout=timeout,
                log_file=log_file,
                log_level=log_level,
                immediate_pause_seconds=immediate_pause_seconds,
            )
            self.delivery = create_delivery(
                resolved_type,
                dconfig,
                db_path=final_db_path,
                poll_interval=poll_interval,
                batch_interval=batch_interval,
                max_attempts=max_attempts,
                max_retries=max_retries,
                max_batch_size=max_batch_size,
                log_bodies=log_bodies,
                raise_on_error=raise_on_error,
            )
        if resolved_type is not None:
            self.ini_manager.set_option(
                "tracker", "AICM_DELIVERY_TYPE", resolved_type.value.upper()
            )

        # Instance-level tracking metadata
        self.customer_key: Optional[str] = None
        self.context: Optional[Dict[str, Any]] = None
        self._anonymize_fields = {str(field) for field in anonymize_fields or []}
        self._anonymizer = anonymizer or self._default_anonymizer

    # ------------------------------------------------------------------
    # Configuration Methods
    # ------------------------------------------------------------------
    def set_customer_key(self, key: str | None) -> None:
        """Update the ``customer_key`` used for tracking."""
        self.customer_key = key

    def set_context(self, context: Dict[str, Any] | None) -> None:
        """Update the ``context`` dictionary used for tracking."""
        self.context = context

    def set_anonymize_fields(self, fields: Iterable[str] | None) -> None:
        """Configure usage payload fields that should be anonymized before sending."""

        self._anonymize_fields = {str(field) for field in fields or []}

    def set_anonymizer(self, anonymizer: Callable[[Any], Any] | None) -> None:
        """Set the callable used to anonymize sensitive values."""

        self._anonymizer = anonymizer or self._default_anonymizer

    # ------------------------------------------------------------------
    # Helper Methods
    # ------------------------------------------------------------------
    def _get_vendor_api_mapping(self, service_key: str) -> Tuple[str, str]:
        """Extract vendor and api_id from service_key.

        Returns:
            Tuple of (vendor, api_id)
        """
        if "::" in service_key:
            vendor = service_key.split("::")[0]
            # Map vendor to api_id
            vendor_to_api = {
                "openai": "openai_chat",
                "anthropic": "anthropic",
                "amazon-bedrock": "amazon-bedrock",
                "fireworks-ai": "fireworks-ai",
                "xai": "openai_chat",  # X.AI uses OpenAI-compatible API
                "google": "gemini",
            }
            api_id = vendor_to_api.get(vendor, vendor)
            return vendor, api_id
        else:
            # If no "::" found, assume it's already an api_id
            return service_key, service_key

    def _build_final_service_key(
        self, service_key: str, api_id: str, response_or_chunk: Any
    ) -> str:
        """Build the final service_key from api_id and model if needed."""
        if "::" not in service_key:
            # service_key was actually an api_id, extract model from response
            model = getattr(response_or_chunk, "model", None)
            if model:
                # Map api_id back to vendor
                api_to_vendor = {
                    "openai_chat": "openai",
                    "openai_responses": "openai",
                    "fireworks-ai": "fireworks-ai",
                    "anthropic": "anthropic",
                    "amazon-bedrock": "amazon-bedrock",
                    "gemini": "google",
                }
                vendor = api_to_vendor.get(api_id, api_id)
                return f"{vendor}::{model}"
        return service_key

    def _resolve_tracking_params(
        self, customer_key: Optional[str], context: Optional[Dict[str, Any]]
    ) -> Tuple[Optional[str], Optional[Dict[str, Any]]]:
        """Resolve tracking parameters using instance defaults as fallbacks."""
        resolved_key = customer_key if customer_key is not None else self.customer_key
        resolved_context = context if context is not None else self.context
        return resolved_key, resolved_context

    def _apply_anonymization(
        self,
        usage: Dict[str, Any],
        *,
        anonymize_fields: Iterable[str] | None,
        anonymizer: Callable[[Any], Any] | None,
    ) -> Dict[str, Any]:
        fields = set(self._anonymize_fields)
        if anonymize_fields:
            fields.update(str(field) for field in anonymize_fields)
        if not fields:
            return usage

        anonymizer_fn = anonymizer or self._anonymizer
        if anonymizer_fn is None:
            return usage

        if not isinstance(usage, Mapping):
            return usage

        return self._anonymize_mapping(usage, fields, anonymizer_fn)

    def _anonymize_mapping(
        self,
        data: Mapping[str, Any],
        fields: set[str],
        anonymizer: Callable[[Any], Any],
    ) -> Dict[str, Any]:
        sanitized: Dict[str, Any] = {}
        for key, value in data.items():
            if key in fields:
                sanitized[key] = self._apply_anonymizer(value, anonymizer)
            else:
                sanitized[key] = self._anonymize_value(value, fields, anonymizer)
        return sanitized

    def _anonymize_value(
        self,
        value: Any,
        fields: set[str],
        anonymizer: Callable[[Any], Any],
    ) -> Any:
        if isinstance(value, Mapping):
            return self._anonymize_mapping(value, fields, anonymizer)
        if isinstance(value, list):
            return [self._anonymize_value(item, fields, anonymizer) for item in value]
        if isinstance(value, tuple):
            return tuple(
                self._anonymize_value(item, fields, anonymizer) for item in value
            )
        if isinstance(value, set):
            return {self._anonymize_value(item, fields, anonymizer) for item in value}
        return value

    def _apply_anonymizer(self, value: Any, anonymizer: Callable[[Any], Any]) -> Any:
        try:
            return anonymizer(value)
        except Exception as exc:  # pragma: no cover - defensive logging
            self.logger.warning("Anonymizer failed: %s", exc)
            return value

    @staticmethod
    def _default_anonymizer(value: Any) -> Any:
        if isinstance(value, str):
            return "[REDACTED]"
        if isinstance(value, bytes):
            return b"[REDACTED]"
        if isinstance(value, bytearray):
            return bytearray(b"[REDACTED]")
        if isinstance(value, Mapping):
            return {
                key: Tracker._default_anonymizer(inner) for key, inner in value.items()
            }
        if isinstance(value, list):
            return [Tracker._default_anonymizer(inner) for inner in value]
        if isinstance(value, tuple):
            return tuple(Tracker._default_anonymizer(inner) for inner in value)
        if isinstance(value, set):
            return {
                f"[REDACTED_{idx}]"
                if isinstance(inner, str)
                else Tracker._default_anonymizer(inner)
                for idx, inner in enumerate(value)
            }
        return value

    def _build_record(
        self,
        service_key: str,
        usage: Dict[str, Any],
        *,
        response_id: Optional[str],
        timestamp: str | datetime | None,
        customer_key: Optional[str],
        context: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        record: Dict[str, Any] = {
            "service_key": service_key,
            "response_id": response_id or uuid4().hex,
            "timestamp": (
                str(timestamp.timestamp())
                if isinstance(timestamp, datetime)
                else str(datetime.fromisoformat(timestamp).timestamp())
                if isinstance(timestamp, str)
                else str(datetime.now(timezone.utc).timestamp())
            ),
            "payload": usage,
        }
        if customer_key is not None:
            record["customer_key"] = customer_key
        if context is not None:
            record["context"] = context
        return record

    # ------------------------------------------------------------------
    # Core Tracking Methods
    # ------------------------------------------------------------------
    def track_batch(
        self,
        records: list[dict],
        *,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> Any:
        """Track multiple usage records in a single batch request.

        This method sends multiple tracking records in a single HTTP request,
        which is more efficient than individual track() calls and avoids
        issues with persistent queue API key mixing.

        Note:
            Maximum batch size is 1000 records. Batches exceeding this limit
            will raise BatchSizeLimitExceeded.

        Args:
            records: List of dictionaries, each containing:
                - service_key (str): The service key for tracking
                - usage (Dict[str, Any]): Usage data to track
                - response_id (Optional[str]): Response ID (auto-generated if None)
                - timestamp (Optional[str | datetime]): Timestamp (current time if None)
                - customer_key (Optional[str]): Customer key (uses instance default if None)
                - context (Optional[Dict[str, Any]]): Context data (uses instance default if None)
            anonymize_fields: Fields to anonymize across all records
            anonymizer: Anonymization function to apply to all records

        Returns:
            For immediate delivery: dict with "results" and "triggered_limits" keys
            For queued delivery: dict with "queued" count

        Example:
            records = [
                {
                    "service_key": "openai::gpt-4",
                    "usage": {"tokens": 100},
                    "customer_key": "customer1",
                    "response_id": "resp1"
                },
                {
                    "service_key": "anthropic::claude-3",
                    "usage": {"tokens": 200},
                    "customer_key": "customer2",
                    "context": {"session": "abc123"}
                }
            ]
            result = tracker.track_batch(records)
        """
        if len(records) > 1000:
            raise BatchSizeLimitExceeded(len(records), 1000)

        if not records:
            return (
                {"results": []} if hasattr(self.delivery, "_enqueue") else {"queued": 0}
            )

        # Build all records first
        built_records = []
        for record_data in records:
            # Extract required fields
            service_key = record_data["service_key"]
            usage = record_data["usage"]

            # Extract optional fields with defaults
            response_id = record_data.get("response_id")
            timestamp = record_data.get("timestamp")
            customer_key = record_data.get("customer_key")
            context = record_data.get("context")

            # Use instance-level values as fallbacks if parameters are None
            resolved_customer_key, resolved_context = self._resolve_tracking_params(
                customer_key, context
            )

            # Apply anonymization to this record's usage
            anonymized_usage = self._apply_anonymization(
                usage, anonymize_fields=anonymize_fields, anonymizer=anonymizer
            )

            # Build the record
            record = self._build_record(
                service_key,
                anonymized_usage,
                response_id=response_id,
                timestamp=timestamp,
                customer_key=resolved_customer_key,
                context=resolved_context,
            )
            built_records.append(record)

        # For immediate delivery, send all records in one batch
        if hasattr(self.delivery, "type") and self.delivery.type.value == "immediate":
            body = {self.delivery._body_key: built_records}
            try:
                data = self.delivery._post_with_retry(body, max_attempts=3)
                if self.delivery._limits_enabled() and isinstance(data, dict):
                    tl_data = data.get("triggered_limits")
                    if tl_data:
                        # Use a ConfigManager that preserves existing INI settings
                        from .config_manager import ConfigManager

                        cfg = ConfigManager(
                            ini_path=self.ini_manager.ini_path, load=True
                        )
                        try:
                            cfg.write_triggered_limits(tl_data)
                        except Exception as exc:  # pragma: no cover
                            self.logger.error("Triggered limits update failed: %s", exc)
                return data
            except Exception as exc:
                self.logger.error("Batch delivery failed: %s", exc)
                raise
        else:
            # For queued delivery, enqueue all records
            total_queued = 0
            response_ids = []
            for record in built_records:
                result = self.delivery.enqueue(record)
                if isinstance(result, int):
                    total_queued = result  # Last queue length
                response_ids.append(record["response_id"])

            return {"queued": total_queued, "response_ids": response_ids}

    def track(
        self,
        service_key: str,
        usage: Dict[str, Any],
        *,
        response_id: Optional[str] = None,
        timestamp: str | datetime | None = None,
        customer_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> Any:
        """Track usage data.

        For immediate delivery, returns a dict with ``result`` and
        ``triggered_limits`` keys. For queued delivery, returns a dict
        ``{"queued": <count>}`` indicating the queue length.
        """
        # Use instance-level values as fallbacks if parameters are None
        customer_key, context = self._resolve_tracking_params(customer_key, context)

        anonymized_usage = self._apply_anonymization(
            usage, anonymize_fields=anonymize_fields, anonymizer=anonymizer
        )

        record = self._build_record(
            service_key,
            anonymized_usage,
            response_id=response_id,
            timestamp=timestamp,
            customer_key=customer_key,
            context=context,
        )
        result = self.delivery.enqueue(record)
        if isinstance(result, int):
            return {"queued": result, "response_id": record["response_id"]}
        return result

    async def track_batch_async(
        self,
        records: list[dict],
        *,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> Any:
        """Async version of track_batch.

        Note:
            Maximum batch size is 1000 records. Batches exceeding this limit
            will raise BatchSizeLimitExceeded.
        """
        if len(records) > 1000:
            raise BatchSizeLimitExceeded(len(records), 1000)

        return await asyncio.to_thread(
            self.track_batch,
            records,
            anonymize_fields=anonymize_fields,
            anonymizer=anonymizer,
        )

    async def track_async(
        self,
        service_key: str,
        usage: Dict[str, Any],
        *,
        response_id: Optional[str] = None,
        timestamp: str | datetime | None = None,
        customer_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> Any:
        return await asyncio.to_thread(
            self.track,
            service_key,
            usage,
            response_id=response_id,
            timestamp=timestamp,
            customer_key=customer_key,
            context=context,
            anonymize_fields=anonymize_fields,
            anonymizer=anonymizer,
        )

    # ------------------------------------------------------------------
    # LLM Tracking Methods
    # ------------------------------------------------------------------
    def track_llm_usage(
        self,
        service_key: str,
        response: Any,
        *,
        response_id: Optional[str] = None,
        timestamp: str | datetime | None = None,
        customer_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> Any:
        """Extract usage from an LLM response and enqueue it.

        Parameters are identical to :meth:`track` except that ``response`` is
        the raw LLM client response.  Usage information is obtained via
        :func:`get_usage_from_response` using the provided ``service_key``.
        ``response`` is returned to allow call chaining. If a ``response_id`` was
        not provided and one is generated, it is attached to the response as
        ``response.aicm_response_id`` for convenience.
        """
        vendor, api_id = self._get_vendor_api_mapping(service_key)

        usage = get_usage_from_response(response, api_id)
        if isinstance(usage, dict) and usage:
            final_service_key = self._build_final_service_key(
                service_key, api_id, response
            )

            track_result = self.track(
                final_service_key,
                usage,
                response_id=response_id,
                timestamp=timestamp,
                customer_key=customer_key,
                context=context,
                anonymize_fields=anonymize_fields,
                anonymizer=anonymizer,
            )
            self._attach_tracking_metadata(response, track_result)
        return response

    def _attach_tracking_metadata(self, response: Any, track_result: Any) -> None:
        """Safely attach tracking metadata to response object."""
        try:
            setattr(
                response,
                "aicm_response_id",
                track_result.get("result", {}).get("response_id"),
            )
            setattr(response, "aicm_track_result", track_result)
        except Exception:
            pass

    async def track_llm_usage_async(
        self,
        service_key: str,
        response: Any,
        *,
        response_id: Optional[str] = None,
        timestamp: str | datetime | None = None,
        customer_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ) -> Any:
        """Async version of :meth:`track_llm_usage`."""
        return await asyncio.to_thread(
            self.track_llm_usage,
            service_key,
            response,
            response_id=response_id,
            timestamp=timestamp,
            customer_key=customer_key,
            context=context,
            anonymize_fields=anonymize_fields,
            anonymizer=anonymizer,
        )

    def track_llm_stream_usage(
        self,
        service_key: str,
        stream: Any,
        *,
        response_id: Optional[str] = None,
        timestamp: str | datetime | None = None,
        customer_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ):
        """Yield streaming events while tracking usage.

        ``stream`` should be an iterable of events from an LLM SDK.  Usage
        information is extracted from events using
        :func:`get_streaming_usage_from_response` and sent via :meth:`track` once
        available.
        """
        vendor, api_id = self._get_vendor_api_mapping(service_key)

        usage_sent = False
        for chunk in stream:
            if not usage_sent:
                usage = get_streaming_usage_from_response(chunk, api_id)
                if isinstance(usage, dict) and usage:
                    # For streaming, try to get model from both stream and chunk
                    chunk_with_model = type(
                        "MockResponse",
                        (),
                        {
                            "model": getattr(stream, "model", None)
                            or getattr(chunk, "model", None)
                        },
                    )()
                    final_service_key = self._build_final_service_key(
                        service_key, api_id, chunk_with_model
                    )

                    self.track(
                        final_service_key,
                        usage,
                        response_id=response_id,
                        timestamp=timestamp,
                        customer_key=customer_key,
                        context=context,
                        anonymize_fields=anonymize_fields,
                        anonymizer=anonymizer,
                    )
                    usage_sent = True
            yield chunk

    async def track_llm_stream_usage_async(
        self,
        service_key: str,
        stream: Any,
        *,
        response_id: Optional[str] = None,
        timestamp: str | datetime | None = None,
        customer_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        anonymize_fields: Iterable[str] | None = None,
        anonymizer: Callable[[Any], Any] | None = None,
    ):
        """Asynchronous version of :meth:`track_llm_stream_usage`."""
        vendor, api_id = self._get_vendor_api_mapping(service_key)

        usage_sent = False
        async for chunk in stream:
            if not usage_sent:
                usage = get_streaming_usage_from_response(chunk, api_id)
                if isinstance(usage, dict) and usage:
                    # For streaming, try to get model from both stream and chunk
                    chunk_with_model = type(
                        "MockResponse",
                        (),
                        {
                            "model": getattr(stream, "model", None)
                            or getattr(chunk, "model", None)
                        },
                    )()
                    final_service_key = self._build_final_service_key(
                        service_key, api_id, chunk_with_model
                    )

                    await self.track_async(
                        final_service_key,
                        usage,
                        response_id=response_id,
                        timestamp=timestamp,
                        customer_key=customer_key,
                        context=context,
                        anonymize_fields=anonymize_fields,
                        anonymizer=anonymizer,
                    )
                    usage_sent = True
            yield chunk

    # ------------------------------------------------------------------
    # Context Manager and Lifecycle Methods
    # ------------------------------------------------------------------
    def __enter__(self):
        """Context manager entry point."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit point - automatically closes the tracker."""
        self.close()

    def close(self) -> None:
        """Release any resources associated with the tracker.

        A ``Tracker`` may be constructed with an explicitly provided delivery
        instance.  Historically such deliveries were treated as externally
        managed and were not stopped when the tracker was closed.  This led to
        surprising behaviour where queued deliveries (like
        :class:`PersistentDelivery`) could continue running after the tracker
        context exited, leaving callers unsure when the queue had fully drained.

        To provide predictable shutdown semantics, ``close()`` now always stops
        the associated delivery instance.  Callers that wish to reuse a single
        delivery across multiple trackers should manage the delivery lifecycle
        separately instead of relying on the tracker's context manager.
        """
        if getattr(self, "delivery", None) is not None:
            self.delivery.stop()



from aicostmanager.ini_manager import IniManager


def test_ini_manager_set_get(tmp_path):
    ini_path = tmp_path / "AICM.INI"
    mgr = IniManager(str(ini_path))
    mgr.set_option("tracker", "AICM_DELIVERY_TYPE", "PERSISTENT_QUEUE")
    assert (
        mgr.get_option("tracker", "AICM_DELIVERY_TYPE") == "PERSISTENT_QUEUE"
    )



import sqlite3
import time

from aicostmanager.delivery.persistent_queue_manager import PersistentQueueManager


def _init_db(path):
    conn = sqlite3.connect(path)
    with conn:
        conn.execute(
            """
            CREATE TABLE queue (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                payload TEXT NOT NULL,
                status TEXT NOT NULL,
                retry_count INTEGER NOT NULL DEFAULT 0,
                scheduled_at REAL NOT NULL,
                created_at REAL NOT NULL,
                updated_at REAL NOT NULL
            )
            """
        )
        now = time.time()
        conn.execute(
            "INSERT INTO queue (payload, status, retry_count, scheduled_at, created_at, updated_at) VALUES ('{}', 'failed', 1, ?, ?, ?)",
            (now, now, now),
        )
    conn.close()


def test_manager_lists_and_requeues_failed(tmp_path):
    db = tmp_path / "queue.db"
    _init_db(str(db))
    mgr = PersistentQueueManager(str(db))
    stats = mgr.stats()
    assert stats.get("failed") == 1
    failed = mgr.list_failed()
    assert len(failed) == 1
    mgr.requeue_failed()
    stats = mgr.stats()
    assert stats.get("failed", 0) == 0



from aicostmanager import CostQueryManager
from aicostmanager.models import CostEventFilters, CostEventItem, CostEventsResponse


class DummyResponse:
    def __init__(self, data):
        self.status_code = 200
        self.ok = True
        self._data = data
        self.headers = {"Content-Type": "application/json"}

    def json(self):
        return self._data


def test_list_costs_typed(monkeypatch):
    monkeypatch.setenv("AICM_API_KEY", "sk")
    manager = CostQueryManager()

    data = {
        "count": 1,
        "next": None,
        "previous": None,
        "results": [
            {
                "provider_id": "v1",
                "service_key": "s1",
                "cost_unit_id": "cu1",
                "quantity": "1",
                "cost_per_unit": "0.01",
                "cost": "0.01",
            }
        ],
    }

    captured = {}

    def requester(self, method, url, **kwargs):
        captured.update(kwargs.get("params", {}))
        return DummyResponse(data)

    monkeypatch.setattr("requests.Session.request", requester)

    filters = CostEventFilters(response_id="resp_123")
    resp = manager.list_costs_typed(filters)

    assert isinstance(resp, CostEventsResponse)
    assert isinstance(resp.results[0], CostEventItem)
    assert resp.results[0].provider_id == "v1"
    assert captured["response_id"] == "resp_123"


def test_iter_costs(monkeypatch):
    monkeypatch.setenv("AICM_API_KEY", "sk")
    manager = CostQueryManager()

    pages = [
        {
            "results": [
                {
                    "provider_id": "v1",
                    "service_key": "s1",
                    "cost_unit_id": "cu1",
                    "quantity": "1",
                    "cost_per_unit": "0.01",
                    "cost": "0.01",
                }
            ],
            "next": manager.client.api_root + "/costs/?offset=1",
        },
        {
            "results": [
                {
                    "provider_id": "v2",
                    "service_key": "s2",
                    "cost_unit_id": "cu2",
                    "quantity": "2",
                    "cost_per_unit": "0.02",
                    "cost": "0.04",
                }
            ],
            "next": None,
        },
    ]

    def requester(self, method, url, **kwargs):
        return DummyResponse(pages.pop(0))

    monkeypatch.setattr("requests.Session.request", requester)

    events = list(manager.iter_costs())
    assert [e.provider_id for e in events] == ["v1", "v2"]
    assert str(events[0].cost) == "0.01"


def test_list_costs_with_context_filters(monkeypatch):
    monkeypatch.setenv("AICM_API_KEY", "sk")
    manager = CostQueryManager()

    captured = {}

    def requester(self, method, url, **kwargs):
        captured.update(kwargs.get("params", {}))
        return DummyResponse(
            {"count": 0, "next": None, "previous": None, "results": []}
        )

    monkeypatch.setattr("requests.Session.request", requester)

    manager.list_costs({"context.environment": "prod"})
    assert captured["context.environment"] == "prod"



import types

from aicostmanager import OpenAIChatWrapper


def test_wrapper_forwards_delivery_type(monkeypatch):
    captured = {}

    class DummyIniManager:
        def get_option(self, section, option, fallback=None):
            return fallback

    class StubTracker:
        def __init__(self, *args, **kwargs):
            captured["delivery_type"] = kwargs.get("delivery_type")
            self.ini_manager = DummyIniManager()

        def track(self, *args, **kwargs):
            pass

        async def track_async(self, *args, **kwargs):
            pass

        def close(self):
            pass

    monkeypatch.setattr("aicostmanager.wrappers.Tracker", StubTracker)

    client = types.SimpleNamespace()
    wrapper = OpenAIChatWrapper(client, delivery_type="PERSISTENT_QUEUE")
    assert captured["delivery_type"] == "PERSISTENT_QUEUE"
    wrapper.close()



from __future__ import annotations

import os
from pathlib import Path

from .utils.ini_utils import atomic_write, file_lock, safe_read_config


class IniManager:
    """Helper for reading and writing generic values to ``AICM.ini``."""

    def __init__(self, ini_path: str | None = None) -> None:
        self.ini_path = self.resolve_path(ini_path)
        with file_lock(self.ini_path):
            self._config = safe_read_config(self.ini_path)

    @classmethod
    def resolve_path(cls, ini_path: str | None = None) -> str:
        """Resolve the path to the INI file."""
        return ini_path or os.getenv(
            "AICM_INI_PATH", str(Path.home() / ".config" / "aicostmanager" / "AICM.INI")
        )

    def _write(self) -> None:
        with file_lock(self.ini_path):
            import io
            content = io.StringIO()
            self._config.write(content)
            atomic_write(self.ini_path, content.getvalue())

    def get_option(self, section: str, option: str, fallback: str | None = None) -> str | None:
        """Return ``option`` from ``section`` or ``fallback`` when missing."""
        with file_lock(self.ini_path):
            config = safe_read_config(self.ini_path)
        if config.has_section(section) and option in config[section]:
            return config[section][option]
        return fallback

    def set_option(self, section: str, option: str, value: str) -> None:
        """Persist ``option`` under ``section`` with ``value``."""
        with file_lock(self.ini_path):
            cfg = safe_read_config(self.ini_path)
            if section not in cfg:
                cfg.add_section(section)
            cfg[section][option] = str(value)
            import io
            buf = io.StringIO()
            cfg.write(buf)
            atomic_write(self.ini_path, buf.getvalue())



"""Python SDK for the AICostManager API."""

__version__ = "0.1.41"

from .client import (
    AICMError,
    APIRequestError,
    AsyncCostManagerClient,
    BatchSizeLimitExceeded,
    CostManagerClient,
    MissingConfiguration,
    NoCostsTrackedException,
    UsageLimitExceeded,
)
from .config_manager import ConfigManager
from .costs import CostQueryManager
from .delivery import (
    Delivery,
    DeliveryConfig,
    DeliveryType,
    ImmediateDelivery,
    PersistentDelivery,
    PersistentQueueManager,
    create_delivery,
)
from .limits import BaseLimitManager, TriggeredLimitManager, UsageLimitManager
from .tracker import Tracker
from .wrappers import (
    AnthropicWrapper,
    BedrockWrapper,
    FireworksWrapper,
    GeminiWrapper,
    OpenAIChatWrapper,
    OpenAIResponsesWrapper,
)

__all__ = [
    "AICMError",
    "APIRequestError",
    "AsyncCostManagerClient",
    "BatchSizeLimitExceeded",
    "CostManagerClient",
    "MissingConfiguration",
    "UsageLimitExceeded",
    "NoCostsTrackedException",
    "ConfigManager",
    "Delivery",
    "DeliveryType",
    "create_delivery",
    "DeliveryConfig",
    "ImmediateDelivery",
    "PersistentDelivery",
    "PersistentQueueManager",
    "Tracker",
    "BaseLimitManager",
    "TriggeredLimitManager",
    "UsageLimitManager",
    "OpenAIChatWrapper",
    "OpenAIResponsesWrapper",
    "AnthropicWrapper",
    "GeminiWrapper",
    "BedrockWrapper",
    "FireworksWrapper",
    "CostQueryManager",
    "__version__",
]
